import enum
from typing import Dict, List, Optional, Set, Union

import numpy as np
from scipy.optimize import brentq


@enum.unique
class VesselHeads(enum.Enum):
    """Defines unique key value pairs to specify the implemented vessel heads."""

    SPHERICAL = "spherical"
    ELLIPSOIDAL = "ellipsoidal 2:1"
    CYLINDRICAL = "flat"
    AMSE_TORISPHERICAL = "amse torispherical"


available_vessel_heads = [head.name for head in VesselHeads]


class Vessel:
    """Class for the calculation of liquid volumes from liquid elevation and vice versa in pressure vessels.

    The heads (or dishes) are approximated as ellipsoidal, specified by the ratio of the semi-minor to the semi-major
    axis. Also supports the use of a weir to divide the vessel in two, where the word weir is a misnomer as it is
    assumed to cover the entire tank diameter.

    Args:
        length (float): length of cylindrical part of vessel
        inner_diameter (float): inner diameter of cylindrical part of vessel
        head_ratio (float, optional):
            Ratio of the semi minor axis to the semi major axis.  Values of C = 1 and C = 0.5
            gives exact partial volumes for spherical and 2:1 ellipsoidal heads,
            respectively. Errors of total volume calculations for AMSE heads are less 0.001%[1].
        head_type (VesselHeads, optional): Specifies the head-type if ``head_ratio`` is not given. See available
            types in inso_toolbox.geometry.vessel.available_vessel_heads. Default ``spherical``
        head_params: (dict, optional): Additional head parameters for non-ellipsoidal heads.

    References:
        [1] : B. Wiencke, Computing the partial volume of pressure vessels,
              https://www.sciencedirect.com/science/article/abs/pii/S0140700709002734
    """

    _required_head_params: Dict[VesselHeads, List[str]] = {
        VesselHeads.SPHERICAL: [],
        VesselHeads.ELLIPSOIDAL: [],
        VesselHeads.CYLINDRICAL: [],
        VesselHeads.AMSE_TORISPHERICAL: ["vessel_wall_thickness"],
    }

    def __init__(
        self,
        length: float,
        inner_diameter: float,
        head_type: VesselHeads = VesselHeads.SPHERICAL,
        head_ratio: Optional[float] = None,
        head_params: Dict[str, float] = {},
    ):
        self._length = length
        self._inner_diameter = inner_diameter

        if head_ratio is None:
            head_ratio = self.get_head_ratio(head_type, inner_diameter=inner_diameter, **head_params)
        self._head_ratio = head_ratio

    @property
    def inner_diameter(self):
        """Inner diameter of vessel"""
        return self._inner_diameter

    @property
    def head_ratio(self):
        """Ratio between semi-major and semi-minor axis of the vessel head when approximated as ellipsoidal."""
        return self._head_ratio

    @property
    def length(self):
        """Length of cylindrical part of vessel, excluding the head."""
        return self._length

    @staticmethod
    def get_head_ratio(head_type: VesselHeads, **head_params) -> float:
        """Calculates the ratio between semi-major axis and semi-minor axis for a given vessel head approximated as an
        ellipsoid.

        Args:
            head_type (VesselHeads): See VesselHeads class for implemented head types.
            **head_params: Additional keyword parameters for different head types. Required arguments:

                * inner_diameter for amse torispherical
                * vessel_wall_thickness for amse torispherical and,
                * inside_knuckle_radius for amse torispherical heads (optional, default = 3 * vessel_wall_thickness)

        Returns:
            ratio (float): ratio between semi-major and semi-minor axis of an ellipsoid approximating the vessel head
        """

        try:
            missing_params: Set[str] = set(Vessel._required_head_params[head_type]) - set(head_params.keys())
        except KeyError:
            raise NotImplementedError(f"Head type {head_type} not implemented")
        if missing_params:
            raise ValueError(f"Head type {head_type} requires parameters {missing_params}")

        if head_type is VesselHeads.SPHERICAL:
            return 1.0
        elif head_type is VesselHeads.ELLIPSOIDAL:
            return 0.5
        elif head_type is VesselHeads.CYLINDRICAL:
            return 0.0
        elif head_type is VesselHeads.AMSE_TORISPHERICAL:
            t = head_params["vessel_wall_thickness"]
            D_i = head_params["inner_diameter"]
            R_k = head_params.get("inside_knuckle_radius", t * 3)
            D_o = t + D_i  # outside diameter
            return 0.30939 + 1.7197 * (R_k - 0.06 * D_o) / D_i - 0.16116 * (t / D_o) + 0.98997 * (t / D_o) ** 2
        else:
            raise NotImplementedError(f"Head type {head_type} not implemented")

    @staticmethod
    def head_volume(h: Union[float, np.ndarray], D_i: float, C: float) -> Union[float, np.ndarray]:
        """
        Approximates the filled volume of a single head as the volume of a semi-ellipsoidal vessel head.

        The parameter C is the between the semi-major and semi-minor axis a and b, respectively, given as C = b / a.
        Corresponding values are defined by AMSE for different head standards. C=1 and C=0.5 corresponds to spherical
        and 2:1 ellipsoidal heads.

        Args:
            h (float or np.ndarray): Height of liquid, 0 < h < D_i. [length]
            D_i (float): Semi-major axis, given by inner diameter of vessel. [length]
            C (float): Ratio between semi-major and semi-minor axis. [unitless]

        Returns:
            The filled volume of the head. [length^3]
        """

        # ensure height is between 0 and diameter
        if np.any(h < 0) or np.any(h > D_i):
            raise ValueError(f"All values of liquid height 'h' must be between 0 and inner diameter D_i = {D_i}")

        # calculate volume
        return D_i ** 3 * C * np.pi / 12 * (3 * (h / D_i) ** 2 - 2 * (h / D_i) ** 3)

    @staticmethod
    def cylinder_volume(h: Union[float, np.ndarray], L: float, D_i: float) -> Union[float, np.ndarray]:
        """
        Calculates the filled volume for a liquid at elevation h inside a a horizontal cylinder of length L and diameter
        D_i, for a liquid at elevation h.  See [1] for derivation.

        Args:
            h (float or np.ndarray): Height of liquid, 0 < h < D_i. [length]
            L (float): Length of cylinder . [length]
            D_i (float): Inner diameter of cylinder. [length]

        Returns:
            The filled volume of the cylinder. [length^3]

        References:
            [1] : Barderas, Stephania, Rodea, How to calculate the volumes of partially full tanks,
                https://www.researchgate.net/publication/301674653_HOW_TO_CALCULATE_THE_VOLUMES_OF_PARTIALLY_FULL_TANKS
        """

        # ensure height is between 0 and diameter
        if np.any(h < 0) or np.any(h > D_i):
            raise ValueError(f"All values of liquid height 'h' must be between 0 and inner diameter D_i = {D_i}")

        R = D_i / 2  # inner radius of vessel

        # calculate volume
        sector_area = L * R ** 2 * np.arccos(1 - h / R)
        triangle_area = L * (R - h) * np.sqrt(2 * R * h - h ** 2)
        return sector_area - triangle_area

    def partially_filled_volume(
        self, elevation: Union[float, np.ndarray], weir_loc: float = 0
    ) -> Union[float, np.ndarray]:
        """
        Calculates the volume of a liquid at a given elevation, possibly with a weir dividing the vessel in two.

        Args:
            elevation (float or np.ndarray): Height of liquid inside vessel
            weir_loc (float, optional): Location of weir, weir_loc = 0 indicates no weir. If present, a single head and
                the cylindrical portion up to the weir is used for volume calculation.

        Returns:
            The filled volume of the vessel (float or np.ndarray)
        """

        # ensure weir is in the cylindrical part of vessel
        if weir_loc < 0 or weir_loc > self.length:
            raise ValueError(f"Expected weir_loc between 0 and length of vessel L = {self.length}, got {weir_loc}")

        effective_length = self.length - weir_loc
        vol_cyl = self.cylinder_volume(elevation, effective_length, self.inner_diameter)
        vol_head = self.head_volume(elevation, self.inner_diameter, self.head_ratio)

        # if no weir, add vol of both heads
        if not weir_loc:
            vol_head *= 2

        return vol_cyl + vol_head

    def full_volume(self, weir_loc: float = 0) -> float:
        """
        Calculates the full volume of the vessel, possibly with a weir dividing the vessel in two.

        Args:
            weir_loc (float, optional): Location of weir, weir_loc = 0 indicates no weir. If present, a single head and
                the cylindrical portion up to the weir is used for volume calculation.
        Returns:
            The full volume of the vessel (float)
        """

        return self.partially_filled_volume(elevation=self.inner_diameter, weir_loc=weir_loc)

    def elevation_from_volume(self, volume: float, weir_loc: float = 0, **minimize_kwargs):
        """
        Calculates the inverse of ``partially_filled_volume``, i.e. calculates the liquid elevation as a function of the
        filled volume.

        A weir boundary may be included at a location given by weir_loc, with which only the volume inside weir_loc is
        used.

        Args:
            volume (float): Volume of the liquid.
            weir_loc (float, optional): Weir location. If present, only a single head and the cylindrical length up to
                the weir location is used for the volume calculation.

        Returns:
            Liquid elevation (float).
        """

        max_volume = self.full_volume(weir_loc=weir_loc)

        if volume < 0 or volume > max_volume:
            raise ValueError(f"Expected volume between 0 and self.full_volume()={max_volume}, got {volume}")

        def minimize_func(h: float, weir_loc: float) -> float:
            return self.partially_filled_volume(h, weir_loc=weir_loc) - volume

        return brentq(minimize_func, args=(weir_loc,), a=0, b=self.inner_diameter, **minimize_kwargs)



import random
from datetime import datetime, timedelta

import numpy as np
import pandas as pd

# random.seed(0)


def create_uniform_data(values: tuple, start_date: datetime = datetime(2020, 7, 23, 15, 27, 0), frequency: str = "1s"):
    values = np.hstack(values)
    return pd.DataFrame({"value": values}, index=pd.date_range(start=start_date, freq=frequency, periods=len(values)))


def create_non_uniform_data(
    values: tuple, start_date: datetime = datetime(2020, 7, 23, 15, 27, 0), time_delta: timedelta = timedelta(seconds=1)
):
    values = np.hstack(values)
    indices = [start_date]
    for i in range(1, len(values)):
        indices.append(indices[i - 1] + time_delta * random.randint(1, 1000))
    return pd.DataFrame({"value": values}, index=indices)


def delete_random_data(data: pd.DataFrame, percentage: float = 0.5) -> pd.DataFrame:
    num_points_to_change = min(int(len(data) * percentage), len(data) - 2)
    indices = np.random.choice(data.index[1:-1], size=num_points_to_change, replace=False)
    return data.drop(indices)


def set_na_random_data(data: pd.DataFrame, percentage: float = 0.5) -> pd.DataFrame:
    new_data = data.copy()
    num_points_to_change = min(int(len(new_data) * percentage), len(new_data) - 2)
    indices = np.random.choice(new_data.index[1:-1], size=num_points_to_change, replace=False)
    new_data.loc[indices] = np.nan
    return new_data


def create_uniform_multidata(
    values: list, start_date: datetime = datetime(2020, 7, 23, 15, 27, 0), frequency: str = "1s"
):
    data = create_uniform_data(values[0], start_date, frequency)
    for i in range(1, len(values)):
        more_data = create_uniform_data(values[i], start_date, frequency)
        m_data = data.merge(more_data, on=data.index)
        m_data.index = data.index
        m_data.drop(columns="key_0", inplace=True)
        data = m_data
    return data



from inso_toolbox.units.units import _UNITS


def list_units_docstring(func):
    """
    Fetches all units and adds them in a list to the docstring.
    """

    def _decorator(func):
        func.__doc__ = "".join([func.__doc__, _UNITS.get_docstring()])
        return func

    return _decorator(func)



from typing import Dict, Union

import pandas as pd
from scipy.signal import savgol_filter

from inso_toolbox.utils import pipeline_function


@pipeline_function
def savitzky_golay(
    data: Union[pd.DataFrame, pd.Series], window_length: int = 9, polyorder: int = 2, optional_kwargs: Dict = {}
) -> Union[pd.DataFrame, pd.Series]:
    """ Savitzky-Golay smoother. Wrapper for scipy.signal.savgol_filter function.
    
    Does not assume uniform time series.

    Args:
        data (pd.DataFrame or pd.Series): Time series to smooth.
        window_length (int): Length of filter window.
        polyorder (int): Order of polynomial used to fit the samples.
        optional_kwargs (Dict): Optional keyword arguments for scipy.signal.savgol_filter.

    Returns:
        pd.DataFrame or pd.Series: Smoothed time series with the same timestamps.
    """  # noqa
    # Cannot contain any NaN values
    if data.isna().any(axis=None):
        raise ValueError("Input time series contains NaN value(s). All missing values must first be imputed.")

    # Apply filter on Series data
    if isinstance(data, pd.Series):
        filtered = savgol_filter(data.values.squeeze(), window_length, polyorder, **optional_kwargs)
        return pd.Series(data=filtered, index=data.index)

    # Otherwise apply filter on each column
    filtered_data = pd.DataFrame()
    for column_name, column_data in data.iteritems():
        filtered_data[column_name] = savgol_filter(
            column_data.values.squeeze(), window_length, polyorder, **optional_kwargs
        )

    filtered_data.index = data.index.values
    return filtered_data



from typing import Callable, List, Optional

from .converter_class import Converter
from .unit_class import Unit
from .unit_group_class import UnitGroup


class Units:
    def __init__(self):
        self._unit_groups = {}
        self._unit_to_group_mapping = {}
        self._group_names = set()
        self._unit_names = dict()
        self._docstring = """"""

    @property
    def _recognized_unit_names(self):
        return self._unit_to_group_mapping.keys()

    def _add_unit_mapping(self, unit_group: UnitGroup) -> None:
        for unit_name in unit_group.recognized_unit_names:
            self._unit_to_group_mapping[unit_name] = unit_group.name

    def add_group(self, group_name: str, units: List[Unit], converters: List[Converter] = None) -> None:
        """
        Method to add a group of units. A group consists of one units and converters between those units.
        Each group must exactly have one SI (default) unit.

        Args:
            group_name (str): A unique name for the unit grouping.
            units (List[Unit]): Units to include in the group. All units added to this class must have unique names
                and synonyms.
            converters (List[Converter]): Converters between units in the group.

        Returns:
            None
        """
        # Check for unique group name
        if group_name in self._group_names:
            raise ValueError(f"Duplicate group name: {group_name}.")
        self._group_names.add(group_name)
        self._docstring += fr"""
        Group: {group_name}

        """

        # Store names of units in the current group
        group_units = set()

        # Check that all unit names and synonyms are unique with all previously added units
        for unit in units:
            equivalent_names = []
            for name in unit:
                if name in self._unit_names:
                    raise ValueError(
                        f"Duplicate unit {name} (exists in both {self._unit_names[name]} and {group_name} unit groups)."
                    )
                self._unit_names[name] = group_name
                group_units.add(name)

                equivalent_names.append(name)

            # Add to docstring
            if unit.is_si_unit:
                self._docstring += fr"""
                * {", ".join(equivalent_names)} (SI Unit)

                """
            else:
                self._docstring += fr"""
                * {", ".join(equivalent_names)}

                """

        # Check that converters convert between units only present in the current group
        if converters is not None:
            for converter in converters:
                if converter.from_unit not in group_units:
                    raise ValueError(
                        "Converting from {} to {} is not possible because {} is not in the {} unit group".format(
                            converter.from_unit, converter.to_unit, converter.from_unit, group_name
                        )
                    )
                if converter.to_unit not in group_units:
                    raise ValueError(
                        "Converting from {} to {} is not possible because {} is not in the {} unit group.".format(
                            converter.from_unit, converter.to_unit, converter.to_unit, group_name
                        )
                    )

        # Assert that there is exactly one SI unit in the group
        if sum(unit.is_si_unit for unit in units) != 1:
            raise ValueError(f"There must be exactly one SI (default) unit in the {group_name} unit group.")

        # Add the unit group to the interal list
        self._add_unit_group(UnitGroup(group_name, units, converters))

    def _add_unit_group(self, unit_group: UnitGroup) -> None:
        self._unit_groups[unit_group.name] = unit_group
        self._add_unit_mapping(unit_group)

    def get_converter(self, from_unit: str, to_unit: Optional[str] = None) -> Callable:
        """Find which UnitGroup to delegate converter search to"""
        # Convert units to lower case
        from_unit = from_unit.lower()
        to_unit = to_unit.lower() if to_unit else None

        # Check if conversion is supported and return converter function
        if from_unit in self._recognized_unit_names and (to_unit in self._recognized_unit_names or not to_unit):
            unit_group_name = self._unit_to_group_mapping.get(from_unit)
            unit_group = self._unit_groups.get(unit_group_name)
            return unit_group.get_converter(from_unit, to_unit)
        else:
            err_msg = f"""
                Conversion from {from_unit} to {to_unit} not supported
                or {from_unit} and/or {to_unit} not a recognized unit.

                Please add the unit or converter to the inso_toolbox.units.units.py file.
            """
            raise ValueError(err_msg)

    def get_docstring(self):
        msg = "Units are placed in groups and can be converted to any other unit within its group. "
        msg += (
            "Each unit group has exactly one default (SI) unit. Most units can be referenced by various names. If there"
        )
        msg += " are any units or unit groups missing please add them to the ``inso_toolbox.units.units.py`` file. "
        msg += "These are the currently existing unit groups:"
        docstring = f"""{msg}

        """
        docstring += self._docstring
        return docstring



from typing import List, Tuple

import numpy as np
import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import DatapointsList, DatapointsQuery

PD_FREQ_HOUR = "h"
PD_FREQ_DAY = "D"
PD_FREQ_WEEK = "W"
PD_FREQ_MONTH_START = "MS"
LEGAL_PERIODIC_FREQS = {
    PD_FREQ_HOUR: "hour",
    PD_FREQ_DAY: "day",
    PD_FREQ_WEEK: "week",
    PD_FREQ_MONTH_START: "month start",
}
WEEK_DAYS = ("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday")


def _verify_and_convert_params_for_periodic_fetch(
    start: int, end: int, freq: str, week_start_day: str
) -> Tuple[pd.Timestamp, pd.Timestamp, str]:
    if freq not in LEGAL_PERIODIC_FREQS:
        raise ValueError(f"Periodic datapoint fetch only supports: {LEGAL_PERIODIC_FREQS}, got '{freq}'")

    if freq == PD_FREQ_WEEK:
        week_start_day = week_start_day.capitalize()
        if week_start_day not in WEEK_DAYS:
            raise ValueError(f"Expected 'week_start_day' to be one of {WEEK_DAYS}, got '{week_start_day}'")
        freq = "-".join((PD_FREQ_WEEK, week_start_day[:3].upper()))  # Match Pandas expectations

    round_freq = freq if freq == PD_FREQ_HOUR else PD_FREQ_DAY
    start = pd.Timestamp(start, unit="ms").floor(round_freq)
    end = pd.Timestamp(end, unit="ms").floor(round_freq)

    return start, end, freq


def _convert_dps_query_result_into_df(
    dpslst_lst: List[DatapointsList], ts_range_int: pd.DatetimeIndex, ts_range: np.ndarray
) -> pd.DataFrame:
    all_dps = [dps for dps_lst in dpslst_lst for dps in dps_lst]  # wut did I just read
    df = pd.concat(
        next((dp.to_pandas().assign(after_ts=ts) for dp in dps if dp.timestamp >= ts_int), pd.DataFrame({"value": []}),)
        for dps, ts_int, ts in zip(all_dps, ts_range_int, ts_range)
    ).dropna()
    return df.assign(lag_ms=(df.index - df["after_ts"]).astype(np.int64) / 1e6)


def fetch_periodic_datapoints(
    client: CogniteClient, external_id: str, start: int, end: int, freq: str, week_start_day: str = "Monday"
) -> pd.DataFrame:
    r"""
    Helper function to get the first raw, non-aggregated datapoint every hour/day/week/month.

    Args:
        client (CogniteClient): Instance of CogniteClient.
        external_id (str): External ID of the time series to fetch.
        start (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970 (UTC), minus leap seconds.
        end (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970 (UTC), minus leap seconds.
        freq (str): Wanted frequency of fetching, e.g.: ``h``: hour, ``D``: day, ``W``: week, ``MS``: month start.
        week_start_day (str): Only used for ``freq=W``, defines what day is considered the first in a week
            (default: ``Monday``).

    Returns:
        pd.DataFrame: Dataframe with three columns; *given external_id* and ``after_ts``, ``lag_ms`` indicating which
        query timestamp produces it - and its lag.

    Warning:
        Using ``freq=MS`` yields a non-fixed frequency, but is also one of the key selling points of this helper
        function(!).

    Hint:
        If you want higher resolution than ``hour``, it is likely that fetching all datapoints, then do the filtration
        yourself is faster

    Example:

        Get first datapoint each week, using Sunday as start of week:

            >>> client = CogniteClient(...)
            >>> first_datapoint_df = fetch_periodic_datapoints(
            ...        client, "foo-bar", start=123, end=456, freq="W", week_start_day="Sunday"
            ... )

    Note:
        This function sends a lot of parallel queries to CDF, so if performance is an issue, consider changing
        ``client.config.max_workers`` before calling (default: 10).
    """
    start, end, freq = _verify_and_convert_params_for_periodic_fetch(start, end, freq, week_start_day)
    ts_range = pd.date_range(start=start, end=end, freq=freq)
    ts_int_range = np.floor_divide(ts_range.asi8, 1e6).astype(int)
    dp_queries = [
        DatapointsQuery(external_id=external_id, start=ts_int, end=ts_int + 1, include_outside_points=True)
        for ts_int in ts_int_range
    ]
    api_results = client.datapoints.query(dp_queries)
    return _convert_dps_query_result_into_df(api_results, ts_int_range, ts_range).rename(columns={"value": external_id})



import inspect
from functools import partial, wraps


def pipeline_function(func):
    """
    Method that makes a function suitable for being used in the Pipeline class.
    The first argument of func must be data: pd.DataFrame.
    This method should be used as a decorator.
    """
    # The first argument of func must be data: pd.DataFrame (set convention in inso-toolbox)
    arguments = inspect.getfullargspec(func).args
    if len(arguments) == 0 or arguments[0] != "data":
        raise TypeError(
            "The first argument of "
            + func.__name__
            + " must be 'data: pd.Dataframe' to be suitable for becoming a pipeline function.\n"
            + "Either remove @pipeline_function decorator or make 'data: pd.Dataframe' the first argument."
        )

    # Set original arguments and docstring
    @wraps(func)
    # Edit behaviour of function based on whether it is used as a function call or for piping
    def piped_function(*args, **kwargs):
        # Return function output if arguments are used or data is passed in
        if len(args) > 0 or "data" in kwargs:
            return func(*args, **kwargs)

        # Otherwise return a function with stored keyword arguments
        else:
            # Check for any invalid keywords (not done automatically by partial)
            invalid_keywords = set(kwargs.keys()).difference(set(arguments))
            if len(invalid_keywords) > 0:
                raise TypeError(
                    func.__name__
                    + " got unexpected keyword arguments: "
                    + ", ".join(str(keyword) for keyword in invalid_keywords)
                )
            # Return function with stored keyword arguments
            return partial(func, **kwargs)

    return piped_function



from typing import Callable, Dict, List, Optional, Tuple, Union

import pandas as pd


class FunctionCall(object):
    def __init__(self, columns, function, output_names=None):
        self.columns = columns
        self.function = function
        self.output_names = output_names

    def __call__(self, data_dict) -> None:
        # Check that the column names exist in the data dictionary
        invalid_columns = set(self.columns).difference(data_dict.keys())
        if len(invalid_columns) > 0:
            raise ValueError("Specified column names do not exist: " + str(invalid_columns))

        # Create dataframe from the selected columns
        input_data = _merge_columns_by_index(self.columns, data_dict)

        # Apply function call, passing in the dataframe
        output_data = self.function(input_data)

        # Check that function returns a pandas dataframe
        if not isinstance(output_data, pd.DataFrame):
            raise TypeError(
                "Output of "
                + self.function.__name__
                + " must be a pandas DataFrame in order to be pipeline compatible."
            )

        # Store updates to be added to data dict
        update_dict = {}

        # Inplace update
        if self.output_names is None:
            for colname in output_data.columns:
                update_dict[colname] = pd.DataFrame(output_data[colname])

        # Not inplace update
        else:
            for i, colname in enumerate(output_data.columns):
                update_dict[self.output_names[i]] = pd.DataFrame(output_data[colname])
                update_dict[self.output_names[i]].columns = [self.output_names[i]]

        # Update data dictionary
        data_dict.update(update_dict)

    def __repr__(self):
        try:
            function_name = str(self.function.func.__name__)
        except AttributeError:
            function_name = str(self.function.__name__)
        if self.output_names is None:
            return "(" + ", ".join(str(col) for col in self.columns) + " -> " + function_name + ")"
        else:
            return (
                "("
                + ", ".join(str(col) for col in self.columns)
                + " -> "
                + function_name
                + " -> "
                + ", ".join(str(name) for name in self.output_names)
                + ")"
            )


class Pipeline(object):
    """
    Pipeline class for chaining together any arbitrary sequence of functions together.

    Example:

        Interpolate data and smooth one of the DataFrame columns::

            >>> from inso_toolbox.imputer import interpolate
            >>> from inso_toolbox.smoother import smoother_sg
            >>> from inso_toolbox import Pipeline
            >>>
            >>> # Let data be a pandas DataFrame with columns "id1" and "id2"
            >>> p = Pipeline()
            >>> p.add_sequential(["id1", "id2"], interpolate(kind="linear", method="average"))
            >>> p.add_sequential("id2", smoother_sg(polyorder=3), "id2_smoothed")
            >>> p.run(data)
    """

    def __init__(self):
        self.queue = []  # Function queue

    def add_sequential(
        self,
        columns: Union[str, List[str]],
        functions: Union[Callable, List[Callable]],
        output_names: Union[str, List[str]] = None,
    ) -> None:
        """
        Specify functions to apply on certain columns present in a DataFrame.

        Args:
            columns (Union[str, List[str]]): Column names for which the functions will be applied on.
            functions (Union[Callable, List[Callable]]): Functions to apply on the specified columns.
                The functions are executed in a sequential manner, one column at a time.
            output_names (Union[str, List[str]]): Final output column names.
                Default behaviour is to overwrite the input data.

        Returns:
            None
        """
        # Clean input data
        columns, functions, output_names = self._clean_add_sequential_input(columns, functions, output_names)

        # Fix inplace operator
        for i, column in enumerate(columns):
            for j, function in enumerate(functions):
                if output_names is None:
                    item = FunctionCall([column], function)
                else:
                    output_name = output_names[i]
                    if j == 0:
                        item = FunctionCall([column], function, [output_name])
                    else:
                        item = FunctionCall([output_name], function)
                self.queue.append(item)

    def run(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Run all functions that have been added to the Pipeline.

        Args:
            data (pd.DataFrame): Data to run the functions on.
                Data column names must match those that have been added to the Pipeline.

        Returns:
            pd.DataFrame
        """
        # Check for valid input
        if not isinstance(data, pd.DataFrame):
            raise TypeError("Input must be a pandas DataFrame")

        # Change data to a dictionary
        data_dict = {}
        for colname in data.columns:
            data_dict[colname] = pd.DataFrame(data[colname])

        # Loop through each FunctionCall, calling the object
        for function_call in self.queue:
            function_call(data_dict)

        # Concatenate the data dictionary back to a DataFrame
        data = _merge_columns_by_index(list(data_dict.keys()), data_dict)

        # Return final data
        return data

    def _clean_add_sequential_input(
        self, columns: Union[str, List], functions: Union[Callable, List], output_names: Union[str, List[str], None],
    ) -> Tuple[List, List, Optional[List]]:
        """
        Returns the cleaned version of columns functions and output_names
        """
        # Convert columns, functions and output_names as lists if not already so
        columns = [columns] if isinstance(columns, str) else columns
        functions = [functions] if callable(functions) else functions
        output_names = [output_names] if isinstance(output_names, str) else output_names

        # Assert that they are all lists now
        for lst, lst_name in zip([columns, functions, output_names], ["columns", "functions", "output_names"]):
            if lst_name == "output_names" and lst is None:
                continue
            if not isinstance(lst, List):
                raise TypeError("Incorrect input type for argument " + lst_name)

        # Check for valid input
        if output_names is not None:
            if len(columns) != len(output_names):
                raise TypeError("Number of output names must match the number of columns")
            if len(set(output_names)) != len(output_names):
                raise TypeError("Output names must be unique")

        # Check that all inserted functions are callable
        for function in functions:
            if not callable(function):
                raise TypeError("Incorrect function call got " + str(function) + " instead of a callable object")

        return columns, functions, output_names

    def __repr__(self):
        return str(self.queue)


def _merge_columns_by_index(columns: List[str], data_dict: Dict) -> pd.DataFrame:
    """
    Convenience method to merge columns together. Raises ValueError if indices don't match.
    """
    data = data_dict[columns[0]]

    for i in range(1, len(columns)):
        colname = columns[i]
        try:
            pd.testing.assert_index_equal(data.index, data_dict[colname].index)
        except AssertionError:
            raise ValueError("All indices must match to merge these time series:\n", columns)
        data = pd.concat((data, data_dict[colname]), axis="columns")

    return data



import warnings
from typing import List, Optional

import pandas as pd
from cognite.client import CogniteClient

from .convert import convert


def convert_from_cdf(
    data: pd.DataFrame,
    client: CogniteClient,
    external_id: str,
    columns: List[str],
    to_unit: Optional[str] = None,
    inplace: bool = False,
) -> pd.DataFrame:
    """
    Convert specified columns in data from given unit in CDF to another unit.
    
    Args:
        data (pd.DataFrame): Data to be converted.
        client (CogniteClient): Client from which the time series unit will be retrieved.
        external_id (str): External id of time series.
        columns (List[str]): List of columns to convert.
        to_unit (str, optional): Specifies which unit the columns should be converted to.
            Default is to convert to SI units.
        inplace (bool, optional): Specifies whether to convert the data inplace or not.
    
    Returns:
        pd.DataFrame: Data with converted columns.
    """  # noqa
    from_unit = client.time_series.retrieve(external_id=external_id).unit

    if from_unit is not None:
        return convert(data, data.columns, from_unit, to_unit, inplace)

    else:
        warnings.warn(f"No unit was found for {external_id}. No unit conversion was done.", RuntimeWarning)
        return data



from typing import Union

import numpy as np

# coming in the next numpy version!
# from numpy.typing import Numerical


Numerical = Union[float, np.ndarray]


def roughness_ratio(eps: Numerical, D: Numerical) -> Numerical:
    """
    Calculates the effective dimensionless roughness parameter eps_r as eps_r = eps / D.

    eps: Pipe roughness [m]
    D:   Pipe diameter  [m]
    """

    if np.any(D <= 0):
        raise ValueError("D must be a positive, non-zero number")

    roughness_r = eps / D
    return roughness_r


def reynolds_number(u: Numerical, rho: Numerical, mu: Numerical, L: Numerical,) -> Numerical:
    """
    Computes the Reynolds number as Re = rho*u*L / mu.

    Args:
        u (float or arraylike): Velocity [m/s]
        rho (float or arraylike): Density [kg/m3]
        mu (float or arraylike): Viscosity [kg/ms]
        L (float or arraylike): Length scale [m]     Often pipe diameter

    Returns:
        Reynolds number (float or arraylike)
    """

    if np.any(mu <= 0):
        raise ValueError("Viscosity mu must be a positive, non-zero number")

    rey_num = rho * u * L / mu
    return rey_num



from typing import List


class Unit:
    """
    Creates a unit. The name and synonymns must be unique from all other unit names.

    Args:
        name (str): Name of the unit. Must be unique.
        synonyms (List(str), optional): Other possible names for the unit. Must all be unique.
        is_si_unit (bool, optioanl): Specifies whether to use this as the default unit.

    Returns:
        None
    """

    def __init__(self, name: str, synonyms: List[str] = None, is_si_unit: bool = False):
        self.name = name
        self.is_si_unit = is_si_unit
        self.synonyms = synonyms if synonyms else []

    def __iter__(self):
        """Iterates over all possible names"""
        yield self.name
        yield from self.synonyms



import warnings
from datetime import datetime
from typing import Union

import numpy as np
import pandas as pd
import scipy.interpolate

from inso_toolbox.utils import pipeline_function
from inso_toolbox.utils.helpers import functional_mean, is_na_all


@pipeline_function
def interpolate(
    data: Union[pd.DataFrame, pd.Series],
    method: str = "linear",
    kind: str = "pointwise",
    granularity: str = "1s",
    start: Union[int, datetime] = None,
    end: Union[int, datetime] = None,
    outside_fill: str = None,
) -> Union[pd.DataFrame, pd.Series]:
    """ 
    Interpolates data for uniform timestamps between start and end with specified frequency.
    
    Args:
        data (pd.DataFrame or pd.Series): Time series to impute.
        method (str): Specifies the interpolation method:

            * 'linear': linear interpolation.
            * 'ffill': forward filling.
            * 'zero', ‘slinear’, ‘quadratic’, ‘cubic’: spline interpolation of zeroth, first, second or third order.
        kind (str): Specifies the kind of returned datapoints:

            * 'pointwise': returns the pointwise value of the interpolated function for each timestamp.
            * 'average': returns the average of the interpolated function within each time period. This is the same as CDF average aggregates.
        granularity (str): Frequency of output e.g. '1s' or '2h'.
        start (int or datetime, optional): Start datetime or timestamp of output dataframe.
        end (int or datetime, optional): End datetime or timestamp of output dataframe.
        outside_fill (str, optional): Specifies how to fill values outside input data range ('NaN' or 'extrapolate').
            Default behaviour is to raise a ValueError if the data range is not within start and end and no outside_fill method is specified.
        
    Returns:
        pd.DataFrame or pd.Series: Uniform, interpolated time series with specified frequency.

    """  # noqa
    # Only pd.Series and pd.DataFrame inputs are supported
    if not isinstance(data, (pd.Series, pd.DataFrame)):
        raise TypeError("Only pd.Series and pd.DataFrame inputs are supported.")

    # Check for empty time series
    if len(data) == 0:
        warnings.warn("The time series is empty.", RuntimeWarning)
        return data

    # Check if all values are NaN
    if is_na_all(data):
        warnings.warn("All values in the time series are NaN.", RuntimeWarning)
        return data

    # Allow for other ways of defining forward filling for stepwise functions
    method = "previous" if method in ("ffill", "stepwise") else method

    # Get outside fill value
    if outside_fill == "NaN":
        outside_fill = np.NaN
    if outside_fill not in (np.nan, "extrapolate") and outside_fill is not None:
        raise TypeError("outside_fill must be either 'NaN' or 'extrapolate'.")

    # Get start and end dates and store as datetime
    if not start:
        start = data.index[0]
    elif isinstance(start, int):
        # If milliseconds provided, recast to float as datetime will throw exception otherwise
        if len(str(start)) == 13:
            start = datetime.fromtimestamp(start / 1000)
        elif len(str(start)) == 10:
            start = datetime.fromtimestamp(start)
        else:
            raise ValueError("Start timestamp is outside valid range.")
    if not end:
        end = data.index[-1]
    elif isinstance(end, int):
        # If milliseconds provided, recast to float as datetime will throw exception otherwise
        if len(str(end)) == 13:
            end = datetime.fromtimestamp(end / 1000)
        elif len(str(end)) == 10:
            end = datetime.fromtimestamp(end)
        else:
            raise ValueError("End timestamp is outside valid range.")
    else:
        start = start

    # Change minute granulatiry to match CDF granularity
    if granularity.lower().endswith("m"):
        granularity = granularity.lower().replace("m", "T")

    # Output timestamps for uniform time series
    timestamps = pd.date_range(start, end, freq=granularity)

    # Only DatetimeIndex is supported
    if not isinstance(data.index, pd.DatetimeIndex):
        raise TypeError("Input data must have DatetimeIndex.")

    # Create uniform x values for output time series
    x_uniform = np.array([timestamp.timestamp() for timestamp in timestamps])

    # Loop over timeseries (univariate Series or multivariate DataFrame)
    series = []
    for ts in pd.DataFrame(data).columns:
        # extract timeseries as pd.Series and drop NaNs
        observations = pd.DataFrame(data)[ts].dropna()

        # x and y datapoints used to construct linear piecewise function
        x_observed = np.array([index.timestamp() for index in observations.index])
        y_observed = observations.values.squeeze()

        # interpolator function
        if outside_fill is None:
            interper = scipy.interpolate.interp1d(x_observed, y_observed, kind=method)
        else:
            interper = scipy.interpolate.interp1d(
                x_observed, y_observed, kind=method, bounds_error=False, fill_value=outside_fill
            )

        # If pointwise, sample directly from interpolated (or original) points
        if kind == "pointwise":
            y_uniform = interper(x_uniform)
        elif kind == "average":
            y_uniform = functional_mean(interper, x_uniform)
        else:
            raise TypeError('kind must be "pointwise" or "average"')

        series.append(pd.Series(data=y_uniform, index=timestamps))

    if isinstance(data, pd.Series):
        return series[0]
    else:
        # dict(zip(.)) recreates original data structure with named columns
        # timestamps as index is already defined by the indivdual Series
        return pd.DataFrame(dict(zip(data.columns, series)))



import math
import warnings
from typing import Union

import numpy as np
from scipy.optimize import root_scalar
from scipy.special import lambertw

Numerical = Union[float, np.ndarray]


def _friction_factor_laminar_dimless(rey_num: Numerical) -> Numerical:
    """
    Friction factor for laminar pipe flow.

    Args:
        rey_num (float or np.ndarray): Reynolds number = rho*u*D/mu

    Returns:
        laminar friction factor (float or np.ndarray) : defined as f_c = 64 / rey_num
    """
    return 64 / rey_num


def _colebrook_solve_dimless(rey_num: float, epsilon_r: float, **kwargs) -> float:
    """
    Colebrook equations for the Darcy friction factor. Solves the implicit colebrook equation using the brentq method.

    Args:
        rey_num (float): Reynolds number = rho*u*D/mu
        epsilon_r (float): Pipe's relative roughness, given as pipe effective roughness (epsilon) scaled with the pipe
            diameter (D) - epsilon_r = epsilon / D
        **kwargs: Extra arguments passed on to scipy.optimize.root_scalar. Can overwrite the default arguments
            given by

            * method = "brentq"
            * xtol = 1.48e-08
            * maxiter = 50

    Returns:
        darcy_friction, f (float or np.ndarray): The Darcy-Weisbach friction factor for a full-flowing circular pipe.
    """

    if np.size(rey_num) > 1 or np.size(epsilon_r) > 1:
        msg = "_colebrook_solve_dimless only available for scalar input."

        # note, this is because scipy does not include vectorized version of a stable root solver.
        raise ValueError(msg)

    f_init = _haaland_dimless(rey_num, epsilon_r)

    def res_colebrook(f, rey_num, epsilon_r):
        sqrtf = math.sqrt(f)
        res = 1 / sqrtf + 2 * math.log10(epsilon_r / 3.7 + 2.51 / (rey_num * sqrtf))
        return res

    # avoid zero-division error
    lower = 1 / (2 * math.log10(max(epsilon_r, 1e-15) / 3.7)) ** 2 / 2
    # ad hoc upper limit
    upper = 1e2

    root_kwargs = dict(xtol=1.48e-08, maxiter=50, method="brentq")
    root_kwargs.update(kwargs)

    res = root_scalar(res_colebrook, bracket=(lower, upper), args=(rey_num, epsilon_r), **root_kwargs)

    if res.converged:
        return res.root
    else:
        warnings.warn("Colebrook equation did not converge, using Haaland value instead.")
        return f_init


def _colebrook_dimless(rey_num: Numerical, epsilon_r: Numerical) -> Numerical:
    """
    Solves the colebrook equations using the Lambert W function.

    Args:
        rey_num (float or np.ndarray): Reynolds number = rho*u*D/mu
        epsilon_r (float or np.ndarray): Pipe's relative roughness, given as pipe effective
            roughness (epsilon) scaled with the pipe diameter (D) - epsilon_r = epsilon / D

    Returns:
        darcy_friction, f (float or np.ndarray): The Darcy-Weisbach friction factor for a full-flowing circular pipe.

    Note:
        Fails for values of Reynolds number > 5x10^8.
    """
    try:
        with np.errstate(divide="raise"):
            a = 2.51 / rey_num
    except (ZeroDivisionError, FloatingPointError) as e:
        raise ValueError("Reynolds number must be non-zero") from e

    b = epsilon_r / 3.7

    argument = math.log(10) / (2 * a) * 10 ** (b / (2 * a))
    denom = 2 * lambertw(argument) / math.log(10) - b / a

    if np.any(denom.imag != 0):
        if np.any(np.isinf(argument)):
            raise ValueError(
                "_colebrook_dimless overflows for large values of reynolds number and/or epsilon_r. "
                "Use another colebrook method"
            )
        else:
            raise ValueError("Numerical error in solving colebrook equation.")

    return (1 / denom ** 2).real


def _colebrook_fast_dimless(rey_num: Numerical, epsilon_r: Numerical) -> Numerical:
    """
    Fast approximation of the colebrook equation, from [1].

    Args:
        rey_num (float or np.ndarray): Reynolds number = rho*u*D/mu
        epsilon_r (float or np.ndarray): Pipe's relative roughness, given as pipe effective
            roughness (epsilon) scaled with the pipe diameter (D) - epsilon_r = epsilon / D

    Returns:
        darcy_friction, f (float or np.ndarray): The Darcy-Weisbach friction factor for a full-flowing circular pipe.

    References:
        [1] "Fast and Accurate Approximations for the Colebrook Equation", Dag Biberg,
            Journal of Fluids Engineering Copyright VC 2017 by ASME MARCH 2017, Vol. 139 / 031401-1,
            https://doi.org/10.1115/1.4034950
    """
    a = 2 / math.log(10)
    b = 2.51
    c = 3.7
    ab = a * b

    try:
        with np.errstate(divide="raise", invalid="raise"):
            log_re_ab = np.log(rey_num / ab)
    except FloatingPointError as e:
        raise ValueError("Reynolds number must be non-zero") from e

    x = log_re_ab + rey_num * epsilon_r / (ab * c)

    g = np.log(x) * (2 / x - 1)
    denominator = a * (log_re_ab + g)
    fric = (1 / denominator) ** 2
    return fric


def _haaland_dimless(rey_num: Numerical, epsilon_r: Numerical) -> Numerical:
    """
    Haaland approximation of the Colebrook equation,
    https://en.wikipedia.org/wiki/Darcy_friction_factor_formulae#Haaland_equation.

    Args:
        rey_num (float or np.ndarray): Reynolds number = rho*u*D/mu
        epsilon_r (float or np.ndarray): Pipe's relative roughness, given as pipe effective
            roughness (epsilon) scaled with the pipe diameter (D) - epsilon_r = epsilon / D

    Returns:
        darcy_friction, f (float or np.ndarray): The Darcy-Weisbach friction factor for a full-flowing circular pipe.
    """
    flow_rough = (epsilon_r / 3.7) ** 1.11
    try:
        with np.errstate(divide="raise"):
            flow_smooth = 6.9 / rey_num
    except (ZeroDivisionError, FloatingPointError) as e:
        raise ValueError("Reynolds number must be non-zero") from e

    try:
        with np.errstate(divide="raise", invalid="raise"):
            denom = -1.8 * np.log10(flow_rough + flow_smooth)
    except FloatingPointError as e:
        raise ValueError("Numerical error in solving colebrook equation, check input.") from e
    flow_rough = (epsilon_r / 3.7) ** 1.11
    flow_smooth = 6.9 / rey_num
    denom = -1.8 * np.log10(flow_rough + flow_smooth)
    f = (1 / denom) ** 2
    return f



import logging

import pandas as pd
from scipy.interpolate import LinearNDInterpolator

logger = logging.getLogger(__name__)


def _check_psv_data(inputs, output, grid):
    # Check to see if output dictionary has too many entries
    if len(output) != 1:
        raise ValueError("outputs dictionary contains too many entries. Should only be 1.")

    # Check if keys for inputs and outputs match grid columns
    keys = set(inputs.keys()).union(set(output.keys()))
    keys_do_not_match = keys - set(grid.columns)
    if keys_do_not_match:
        raise KeyError("input keys do not match grid column names")

    # Check if input values are outside the bounds of the grid
    for key, val in inputs.items():
        max_val = max(grid[key])
        min_val = min(grid[key])
        if max_val < val < min_val:
            raise ValueError(f"Input value for {key}:{val} is out of bounds min: {min_val}, max: {max_val}")


def psv_utilization(inputs, output, grid: pd.DataFrame, obey_threshold=False, max_psv_util=200.0):
    """
    Generic method for calculating pressure safety valve (PSV) utilization
    input and output keys must match the column names in the grid

    Args:
        inputs (dict): Independent variables, must match column names of grid
        output (dict): Dependent variable, must match the column name of grid
        grid (pd.DataFrame): PSV grid
        obey_threshold (bool, optional): Specifies whether a threshold is placed on the calculation. Default is False.
        max_psv_util (float, optional): Threshold value. Default if 200%.

    Returns:
        float: PSV utilization (percentage 0-100)
    """
    _check_psv_data(inputs, output, grid)

    # Generate the interpolator
    try:
        x = grid[inputs.keys()]
        y = grid[output.keys()]
        f = LinearNDInterpolator(x, y)
    except Exception as e:  # Raise errors due to LinearNDInterpolator
        logger.error("Unable to generate PSV grid interpolation function.")
        raise e

    # Calculate the actual utilization
    actual_util = f(list(inputs.values()))
    max_util = list(output.values())[0]
    psv_util = actual_util / max_util * 100
    logger.debug(f"actual util = {actual_util}, max util = {max_util}, PSV utilization = {psv_util}")

    # If PSV value is higher than specified threshold, give error message and return threshold instead of value.
    if obey_threshold:
        if psv_util > max_psv_util:
            logger.warning(f"PSV utlization is very high ({psv_util})! Returning {max_psv_util}%")
            return max_psv_util
        else:
            return psv_util
    else:
        return psv_util



import logging

import numpy as np

logger = logging.getLogger(__name__)


# Base physics calculations needed for rhov2


def mass_rate_liq(rate, density):
    """
    Method to calculates the mass of a liquid given rate and density
    Args:
        rate (float): rate (m3/s)
        density (float): density (kg/m3)
    Returns:
        float: mass rate (kg/s)
    """
    mass_rate = rate * density
    return mass_rate


def mass_rate_gas(rate, M=22.4, z_std=0.9956, p_std=101.325, t_std=288):
    """
    Method to calculates the mass of a gas given rate (mmscfd), based on ideal gas law
    Args:
        rate (float): gas rate (m3/s)
        M (float, optional): Molar mass (kg/mol)
        z_std (float, optional): z-factor at standard conditions (unitless)
        p_std (float, optional): pressure at standard conditions (kPa)
        t_std (float, optional): temperature at standard conditions (Kelvin)
    Returns:
        float: mass rate in kg/s
    """
    R = 8.314  # universal gas constant, J / mol·K
    mass_rate = rate * p_std * M / (z_std * t_std * R)
    return mass_rate  # kg/s


def velocity(rate: float, nozz_id: float):
    """
    Method to calculate velocity given a rate
    Args:
        rate (float): fluid rate (m3/s)
        nozz_id (float): nozzle inner diameter (mm)
    Returns:
        float: velocity in m/s
    """
    nozz_area = np.pi / 4 * (nozz_id / 1000) ** 2  # m2
    return rate / nozz_area  # m/s


def fluid_momentum(density: float, velocity: float):
    """
    Standard equation for calculating fluid momentum
    Ref: https://petrowiki.spe.org/Separator_sizing#Nozzle_sizing
    Args:
        density (float): density of the fluid (kg/m3)
        velocity (float): velocity of the fluid (m/s)
    Returns:
        float: fluid momentum in Newton (N)
    """
    return density * velocity ** 2



from typing import Callable, List, Tuple, Union

import numpy as np
import pandas as pd
import scipy.integrate


def check_uniform(data: pd.DataFrame, unit: str = "ns") -> bool:
    """
    Convenience method to verify input time series is uniform.
    Args:
        data (pandas.DataFrame): Time series to check. Object must have a datetime-like index.
        unit (str): Time unit of uniformity check. Follows Numpy DateTime Units convention.
    Returns:
        bool
    """
    # Get time difference between successive observations
    timediff = np.diff(data.index)
    # Round to desired time unit (e.g. uniform on second scale)
    rounded = timediff.astype(dtype=f"timedelta64[{unit}]")
    # Check whether uniform
    return len(np.unique(rounded)) == 1


def make_uniform(data: pd.DataFrame, resolution: str = "1s", interpolation: str = None) -> pd.DataFrame:
    """
    Convenience method to make an input time series dataframe uniform per specified temporal resolution.
    Object must have a datetime-like index.
    Args:
        resolution (str):  Temporal resolution of uniform time series. Follows Pandas DateTime convention.
        interpolation (optional str): Method for optional interpolation. Follows pandas Resampler.interpolate
    Returns:
        pandas.DataFrame
    """
    new_data = data.copy()
    # If no interpolation specified, do not perform interpolation
    if not interpolation:
        new_data = new_data.resample(resolution).mean()
    # Otherwise perform desired interpolation
    else:
        new_data = new_data.resample(rule=resolution).mean().interpolate(method=interpolation)
    return new_data


def functional_mean(function: Callable, x_vals: List) -> List:
    """
    Convenience method to calculate the mean of a function inbetween each x value.
    The last point is the function called at the last x value. This is required for the
    input and output lists to have the same lengths.
    """
    assert len(x_vals) > 0
    y_vals = np.zeros(len(x_vals))
    for i in range(len(x_vals) - 1):
        y_vals[i] = scipy.integrate.quad(function, x_vals[i], x_vals[i + 1])[0] / (x_vals[i + 1] - x_vals[i])
    y_vals[len(x_vals) - 1] = function(x_vals[-1])
    return y_vals


def is_na_all(data: Union[pd.DataFrame, pd.Series]) -> bool:
    """
    Convenience method to test if all values in dataframe/series are NaN.
    """
    if isinstance(data, pd.Series):
        return data.isnull().all()
    elif isinstance(data, pd.DataFrame):
        return pd.isna(data).all(axis=None)
    else:
        raise ValueError("Convenience method only supports Series or DataFrame.")


def is_na_initial(data):
    """
    Convenience method to test if any initial values in dataframe/series are NaN.
    """
    if isinstance(data, pd.Series):
        return pd.isna(data.iloc[0])
    elif isinstance(data, pd.DataFrame):
        return pd.isna(data.iloc[0]).any(axis=None)
    else:
        raise ValueError("Convenience method only supports Series or DataFrame.")


def is_na_final(data):
    """
    Convenience method to test if any final values in dataframe/series are NaN.
    """
    if isinstance(data, pd.Series):
        return pd.isna(data.iloc[-1])
    elif isinstance(data, pd.DataFrame):
        return pd.isna(data.iloc[-1]).any(axis=None)
    else:
        raise ValueError("Convenience method only supports Series or DataFrame.")


def cross_corr(x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Computes between signal x and y correlation for different lags and returns corresponding lag and correlation
    coefficients

    Args:
        x (np.ndarray): Signal 1
        y (np.ndarray): Signal 2

    Returns:
        (np.ndarray, np.ndarray): Lags and correlation coefficients

    """
    x, y = map(np.asarray, (x, y))
    n = x.size  # think n is better than lx which refers to one of the two variables
    if n != y.size:
        raise ValueError("x and y must be equal length")

    lags_num = n - 1

    if lags_num >= n or lags_num < 1:
        raise ValueError(f"lags must be None or strictly positive < {n}")

    corr_coeffs = np.correlate(x, y, mode="full")

    lags = np.arange(-lags_num, lags_num + 1)
    start_idx, end_idx = n - 1 - lags_num, n + lags_num
    corr_coeffs = corr_coeffs[start_idx:end_idx]

    return lags, corr_coeffs



"""
Functions used to visualize data from Oscillations Data Quality Metrics
"""
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec


def plot_lpc_roots(
    threshold: float,
    time: np.ndarray,
    data: np.ndarray,
    predicted: np.ndarray,
    f: np.ndarray,
    pxx: np.ndarray,
    roots: np.ndarray,
    distances: np.ndarray,
    figsize: Tuple = (9, 6),
) -> Tuple[plt.Figure, list]:
    """

    Args:
        threshold (float):  Threshold Radius of the circle to determine if a signal has oscillatory components
        time (np.ndarray): Time vector of the signal
        data (np.ndarray): Raw data vector
        predicted (np.ndarray): Predicted polynomial obtained form the LPC analysis
        f (np.ndarray): Frequency vector from the power spectral density analysis
        pxx (np.ndarray): Power spectral density of the signal
        roots (np.ndarray): Roots of the LPC coefficient
        distances (np.ndarray): Distance to the unit circle for each root
        figsize (Tuple): Size of figure

    Returns:

        Tuple[plt.Figure, list]
    """

    fig = plt.figure(constrained_layout=True, figsize=figsize)

    gs = GridSpec(2, 3, figure=fig)
    ax1 = fig.add_subplot(gs[0, :])
    ax2 = fig.add_subplot(gs[1, :-1])
    ax3 = fig.add_subplot(gs[1, 2])

    ax1.plot(time, data, marker=".", label="Raw")
    ax1.plot(time, predicted, "-", label="LPC prediction")
    ax1.set_xlabel("Time")
    ax1.set_ylabel("Amplitude")
    ax1.set_title("Signal")
    ax1.legend()

    ax2.loglog(f, pxx)
    ax2.set_xlabel("Frequency")
    ax2.set_ylabel("Amplitude")
    ax2.set_title("Power Spectral Density")

    # Unit circle and oscillation threshold
    # If there are roots inside the threshold the signal present signal
    # A root with a distance to unit circle less than 0.2 is label as data with signal
    ang = np.linspace(0, np.pi * 2, 100)
    x_unit = np.cos(ang)
    y_unit = np.sin(ang)
    x_threshold = threshold * np.cos(ang)
    y_threshold = threshold * np.sin(ang)
    # For filling area in plot
    xf = np.concatenate((x_unit, x_threshold[::-1]))
    yf = np.concatenate((y_unit, y_threshold[::-1]))
    ax3.plot(x_unit, y_unit, linewidth=1, color="r")
    ax3.plot(x_threshold, y_threshold, linewidth=1, color="r")
    ax3.axvline(x=0, color="k", linewidth=0.5)
    ax3.axhline(y=0, color="k", linewidth=0.5)
    # Plot roots
    for root, distance in zip(roots, distances):
        if distance > (1 - threshold):
            ax3.plot(root.real, root.imag, "ro", markersize=7)
        else:
            ax3.plot(root.real, root.imag, "go", markersize=7)

    ax3.set_xlabel("Real Part")
    ax3.set_ylabel("Imaginary Part")
    ax3.set_title("LPC Roots in z-plane")

    ax3.fill_between(xf, yf, color="r", alpha=0.05)
    ax3.set_aspect("equal", adjustable="box")

    return fig, [ax1, ax2, ax3]


def plot_peaks(
    lags: np.ndarray,
    xcoef: np.ndarray,
    freqs: np.ndarray,
    power_norm: np.ndarray,
    peak_freqs: np.ndarray,
    peak_power_norm: np.ndarray,
    figsize: Tuple = (9, 4),
) -> Tuple[plt.Figure, plt.Axes]:
    """

    Plots lags and correlation coefficients and peak frequencies.

    Args:
        lags (np.ndarray): Lags from cross correlation computations
        xcoef (np.ndarray): Correlation coefficients corresponding to the lags
        freqs (np.ndarray): Frequencies
        power_norm (np.ndarray): Normalized power coresponding to frequencies in freqs
        peak_freqs (np.ndarray): Detected peak prequencies
        peak_power_norm (np.ndarray): Normalized power of peak frequencies
        figsize (Tuple): size of figure

    Returns:
        Tuple[plt.Figure, plt.Axes]

    """
    fig, ax = plt.subplots(1, 2, figsize=figsize)
    ax[0].plot(lags, xcoef)
    ax[0].set_title("Correlation coefficient with positive lags")
    ax[0].set_xlabel("Lags")
    ax[0].set_ylabel("Correlation coefficients")
    ax[1].plot(freqs, power_norm)
    ax[1].axhline(y=0.3, color="r", linestyle="--")
    ax[1].set_title("Normalized FFT magnitude spectrum")
    ax[1].plot(peak_freqs, peak_power_norm, "go", markersize=8, alpha=0.5)
    ax[1].set_xlabel("Frequency")
    fig.tight_layout()

    return fig, ax



import warnings
from typing import Dict, Union

import pandas as pd
import scipy.signal as signal


def butterworth(data: Union[pd.DataFrame, pd.Series], butter_kwargs: Dict = {}) -> Union[pd.DataFrame, pd.Series]:
    """ Butterworth filter.
    Uses second-order section (sos) representation, polynomial (ba) representation,
    or zeros, poles, and system gain (zpk) representation.

    Args:
        data (pd.DataFrame or pd.Series): Time series to filter.
        butter_kwargs (Dict): Keyword arguments for scipy.signal.butter filter.

    Returns:
        pd.DataFrame or pd.Series: Filtered signal.
    """  # noqa

    # Only pd.Series and pd.DataFrame inputs are supported
    if not isinstance(data, pd.Series) and not isinstance(data, pd.DataFrame):
        raise ValueError("Only pd.Series and pd.DataFrame inputs are supported.")

    # Warn user if there are any missing values
    if data.isna().any(axis=None):
        warnings.warn("There are missing values present in the time series.", RuntimeWarning)

    if butter_kwargs.get("output") == "sos":
        # create filter
        filter_output = signal.butter(**butter_kwargs)
        # apply filter
        filtered = signal.sosfilt(filter_output, data, axis=0)
    elif butter_kwargs.get("output") == "ba":
        # create filter
        b, a = signal.butter(**butter_kwargs)
        # apply filter
        filtered = signal.lfilter(b, a, data, axis=0)
    elif butter_kwargs.get("output") == "zpk":
        # create filter
        z, p, k = signal.butter(**butter_kwargs)
        # Return polynomial transfer function representation from zeros and poles
        b, a = signal.zpk2tf(z, p, k)
        # apply filter
        filtered = signal.lfilter(b, a, data, axis=0)

    # Return series if that was the input type
    if isinstance(data, pd.Series):
        return pd.Series(filtered, index=data.index)

    # Return dataframe with same timestamps
    return pd.DataFrame(data=filtered, index=data.index, columns=data.columns)



from typing import Callable


class Converter:
    """
    Creates a converter function between two units.

    Args:
        from_unit (str): Name of unit to convert from.
        to_unit (str): Name of unit to convert to.
        function (Callable): Function that converts from_unit to to_unit.
            The function must have one input argument and return a scalar.

    Returns:
        None
    """

    __slots__ = ["from_unit", "to_unit", "function"]

    def __init__(self, from_unit: str, to_unit: str, function: Callable):
        self.from_unit = from_unit
        self.to_unit = to_unit
        self.function = function



from typing import Callable, List, Optional

from .converter_class import Converter
from .unit_class import Unit


class UnitGroup:
    def __init__(self, name: str, units: List[Unit], converters: List[Converter] = None):
        self.name = name.upper()
        self.units = units
        self.converters = converters  # type: ignore

        if not self.si_unit:
            raise ValueError(f"No SI unit specified for {self.name}")

    @property
    def si_unit(self) -> Unit:
        si_units = [unit for unit in self.units if unit.is_si_unit]

        if len(si_units) != 1:
            raise ValueError(f"Exactly one of the {self.name} units has to be marked as a SI unit.")

        return si_units[0]

    @property
    def converters(self) -> List[Converter]:
        return self._converters

    @converters.setter
    def converters(self, converters: List[Converter]):
        if converters:
            unit_names = [unit.name for unit in self.units]
            for converter in converters:
                if not (converter.from_unit in unit_names and converter.to_unit in unit_names):
                    err_msg = f"""
                        Cannot find both [{converter.from_unit}] and [{converter.to_unit}]
                        among the {self.name} units: {unit_names}
                    """
                    raise ValueError(err_msg)
        else:
            converters = []
        self._converters = converters

    @property
    def unit_names(self):
        return [unit.name for unit in self.units]

    @property
    def recognized_unit_names(self):
        """Unit names and synonyms"""
        all_synonyms = []
        for unit in self.units:
            all_synonyms.extend(unit.synonyms)
        return self.unit_names + all_synonyms

    def __iter__(self):
        yield from self.units

    def find_unit_name(self, name: Optional[str] = None) -> str:
        for unit in self.units:
            if name in unit:
                return unit.name

        raise ValueError(f"Unit {name} not recognized.")

    def get_converter(self, from_unit: str, to_unit: Optional[str] = None) -> Callable:
        if not to_unit:
            to_unit = self.si_unit.name
        return self._get_converter(from_unit, to_unit)

    def _get_converter(self, from_unit: str, to_unit: Optional[str] = None) -> Callable:
        """Returns None if conversion is from the """
        if from_unit not in self.unit_names:
            from_unit = self.find_unit_name(from_unit)

        if to_unit not in self.unit_names:
            to_unit = self.find_unit_name(to_unit)

        if from_unit != to_unit:
            # Find and return the converter between the two units
            for converter in self.converters:
                if from_unit == converter.from_unit and to_unit == converter.to_unit:
                    return converter.function
        else:
            # If conversion to and from same unit
            return lambda x: x

        raise ValueError(f"Conversion from {from_unit} to {to_unit} not supported.")



import warnings
from typing import Dict, Union

import pandas as pd
from scipy import signal

from inso_toolbox.utils import pipeline_function


@pipeline_function
def chebyshev(
    data: Union[pd.DataFrame, pd.Series], filter_type: int = 1, cheby_kwargs: Dict = {}
) -> Union[pd.DataFrame, pd.Series]:
    """ Chebyshev filter of type 1 and 2.
    Uses second-order section (sos) output filter type.

    Args:
        data (pd.DataFrame or pd.Series): Time series to filter.
        filter_type (int): Chebyshev filter type (either 1 or 2).
        cheby_kwargs (Dict): Keyword arguments for scipy.signal.cheby filter (must include "output":"sos").

    Returns:
        pd.DataFrame or pd.Series: Filtered signal.
    """  # noqa
    # Only pd.Series and pd.DataFrame inputs are supported
    assert isinstance(data, pd.Series) or isinstance(data, pd.DataFrame)

    # Only type 1 and 2 chebyshev filters exist
    if filter_type not in {1, 2}:
        raise ValueError("Filter type must be either 1 or 2.")

    # Scipy also supports output types "ab" and "zpk". Here only "sos" is supported.
    if cheby_kwargs.get("output") != "sos":
        raise ValueError("Only sos output filter type is supported.")

    # Warn user if there are any missing values
    if data.isna().any(axis=None):
        warnings.warn("There are missing values present in the time series.", RuntimeWarning)

    # Get type 1 and 2 chebyshev filter
    cheby_filters = {1: signal.cheby1, 2: signal.cheby2}

    # Get filter output
    filter_output = cheby_filters[filter_type](**cheby_kwargs)

    # Filter the data
    filtered = signal.sosfilt(filter_output, data, axis=0)

    # Return series if that was the input type
    if isinstance(data, pd.Series):
        return pd.Series(filtered, index=data.index)

    # Return dataframe with same timestamps
    return pd.DataFrame(data=filtered, index=data.index, columns=data.columns)



from typing import Dict, Tuple, Union

import numpy as np
import pandas as pd
from numba import njit
from scipy.fft import fft, fftfreq
from scipy.signal import argrelextrema, lfilter

from inso_toolbox.detectors.utils.visualization import plot_lpc_roots, plot_peaks
from inso_toolbox.utils.helpers import cross_corr


def oscillation_detector(
    data: Union[pd.DataFrame, pd.Series], order: int = 4, threshold: float = 0.2, visualize: bool = False
) -> Dict:

    """
        Identifies if a signal contains one or more oscillatory components. Based on the paper by Sharma et al. [#]_

            .. [#] Sharma et al.
               Automatic signal detection and quantification in process control loops using linear predictive coding.
               Eng. Sc. & Tech. an Intnl. Journal
               2020.

        The method uses Linear Predictive Coding (LPC) and is implemented as a 3 step process:

            1. Estimate the LPC coefficients from the prediction polynomial. These are used to estimate a fit to the
               data
            2. Estimate the roots of the LPC coefficients
            3. Estimate the distance of each root to the unit circle in the complex plane
        If the distance of any root is close to the unit circle (less than 0.2) the signal is considered to have an
        oscillatory component


        Args:
            data (Union[pd.DataFrame, pd.Series]): Data to detect signal on. Index must be pd.DatetimeIndex or integer
             and dataframe can only contain one column in addition to index.
            order (int, optional): Order of the prediction polynomial.
            threshold (float, optional): Maximum distance of a root to the unit circle for which the signal is
             considered to have an oscillatory component.
            visualize (bool, optional): If True results are plotted, otherwise no plot is generated.


        Returns:
            dict: Dictionary with the following keys and values::

                {
                    "roots": np.ndarray,
                    "distances": np.ndarray,
                    "PSD": dict:  {"f": np.ndarray, "Pxx": np.ndarray},
                    "fit": dict: {"time": np.ndarray, "data": np.ndarray},
                    "oscillations": bool,
                    "lpc_fig": dict: {"fig": plt.Figure , "axes": list,},
                    "peaks_fig": dict: {"fig": plt.Figure , "ax": plt.Axes,},
                    "peaks": dict: {"f": np.ndarray, "amplitude": np.ndarray,}
                }

            Return dictionary elaboration:

            - roots -> roots of the predicted LPC coefficients
            - distances -> distance of each root to the unit circle
            - PSD -> Power spectral density, frequency and power vector
            - fit -> fitted data using the LPC prediction polynomial
            - oscillations -> (1) Oscillation detected, (0) no oscillatory component detected
            - lpc_fig -> Figure handle used for visualizing LPC data.
              If it is not initialized/requested it is an empty dict
            - peaks_fig -> Figure handle used for visualizing detected peak frequencies data.
              If it is not initialized/requested it is an empty dict.
            - peaks -> Peak frequencies and corresponding amplitudes in original data.

        Warnings:
            Large variations in sampling time may affect the proficiency of the algorithm.


        Example:
            Detection of oscillatory components and plotting of results.

            >>> from cognite.client import CogniteClient
            >>> c = CogniteClient()
            >>> df = c.datapoints.retrieve_dataframe(id=[1], start="2w-ago", end="now",
                ...         aggregates=["average"], granularity="1m", complete="fill,dropna")
            >>> results = oscillation_detector(df, visualize=True)

    """

    if isinstance(data, pd.Series):
        pass
    elif isinstance(data, pd.DataFrame):
        if len(data.columns) != 1:
            raise ValueError("Input data must have exactly one column!")
        data = data.squeeze()
    else:
        raise TypeError("Expected 'pd.Series' or 'pd.DataFrame', got {type(data)}!")

    time = data.index.astype(np.int64)
    data = data.to_numpy()

    # Resample with frequency equal to average frequency into uniformly sampled data
    # In most asset-heavy industrial data, the sampling frequency
    # is non-uniform and sampling intervals range from a couple of seconds to minutes
    # In some cases the sampling interval can be longer than minutes. In that case the method could fail

    delta_t = np.median(np.diff(time))  # Sampling interval
    time_interp = np.arange(time[0], time[-1], delta_t)
    data_interp = np.interp(time_interp, time, data)

    # Estimate LPC and generate prediction polynomial
    lpc_arr = lpc(data_interp, order)  # First coefficient is always 1
    data_predicted = lfilter(-lpc_arr[1:], [1], data_interp)  # Predicted data fit

    if len(data_interp) != len(data_predicted):
        raise RuntimeError(
            f"Something went wrong! The length of the interpolated ({len(data_interp)}) and predicted data"
            f" ({len(data_predicted)}) are different"
        )

    # Replace initial output from the filter (outliers)
    # This is to avoid large deviations in the initial prediction introduced by the filter
    data_interp_std = np.std(data_interp)
    for i, (pred, interp) in enumerate(zip(data_predicted, data_interp)):
        if abs(pred - interp) > data_interp_std:
            data_predicted[i] = interp
        else:
            break

    roots = np.roots(lpc_arr)
    distances = 1 - abs(roots)

    # Verify if signal were detected. Any distance below given threshold (close to the unit circle. default 0.2)
    # means that an oscillatory component was detected
    detected = any(distances < threshold)

    # Obtain power spectral density of signal
    n = len(data_interp)
    data_interp_fft = fft(data_interp)
    amplitude = 2 * np.abs(data_interp_fft)[: n // 2] / n
    data_freqs = fftfreq(n, delta_t)
    data_freqs = data_freqs[: n // 2]

    # Build results dictionary and plot if requested
    results = {
        "roots": roots,
        "distances": distances,
        "PSD": {"f": data_freqs, "Pxx": amplitude},
        "fit": {"time": time_interp, "data": data_predicted},
        "oscillations": detected,
        "lpc_fig": {},
        "peaks": {},
        "peaks_fig": {},
    }
    if visualize:
        _visualize_lpc_analysis(
            threshold, time_interp, data_interp, data_predicted, data_freqs, amplitude, roots, distances, results
        )

    # If oscillations detected find amplitude of peak frequencies, and plot if requested.
    if detected:
        residuals = data_interp - data_predicted
        peaks, peaks_fig = _peak_freq_components(residuals, data_interp, delta_t, visualize)

        results["peaks"] = peaks
        results["peaks_fig"] = peaks_fig

    return results


def _visualize_lpc_analysis(
    threshold: float,
    time: np.ndarray,
    data_interp: np.ndarray,
    data_pred: np.ndarray,
    data_freqs: np.ndarray,
    amplitude: np.ndarray,
    roots: np.ndarray,
    distances: np.ndarray,
    results: Dict,
) -> Dict:
    radius = 1 - threshold
    plot_time_interp = pd.to_datetime(time)
    fig, lpc_axes = plot_lpc_roots(
        radius, plot_time_interp, data_interp, data_pred, data_freqs, amplitude, roots, distances
    )
    lpc_fig = {"fig": fig, "axes": lpc_axes}
    results["lpc_fig"] = lpc_fig
    return results


def _peak_freq_components(residuals: np.ndarray, data: np.ndarray, dt: float, visualize: bool) -> Tuple[Dict, Dict]:
    """
    Find the peak frequency components of a signal identified as having oscillation via the LPC method.

    Args:
        residuals (np.ndarray): residuals from the predicted LPC analysis.
        data (np.ndarray): signal. It's assumed the data is uniformly sampled (constant sampling interval)
        dt (float): sampling interval.
        visualize (bool): If True results are plotted, otherwise no plot is generated.
    Returns:
        (dict, dict): tuple of two dictionaries where the first contain peak frequencies and amplitudes and the second
                      contains figure components
    """
    lags, xcoef = cross_corr(residuals, data)

    # Remove negative lags and subtract its mean
    mask = lags >= 0
    xcoef = xcoef[mask]
    lags = lags[mask]
    xcoef -= np.mean(xcoef)

    # Compute FFT of correlation coefficients and normalize with max peak
    power_norm = np.abs(fft(xcoef))
    power_norm /= np.max(power_norm)
    n = len(power_norm)
    freqs = fftfreq(n, dt)

    # Remove half of the results as they are the mirrored
    power_norm = power_norm[: n // 2]
    freqs = freqs[: n // 2]

    # Locate peak frequencies
    peaks_loc = argrelextrema(np.where(power_norm > 0.3, power_norm, [0]), np.greater)[0]
    peak_freqs = freqs[peaks_loc]
    peak_power_norm = power_norm[peaks_loc]

    # FFT spectrum of original signal
    n1 = len(data)
    data_fft = fft(data)
    amplitude = 2 * np.abs(data_fft)[: n1 // 2] * 1 / n1
    data_freqs = fftfreq(n1, dt)
    amplitude_index = [np.where(data_freqs[: n1 // 2] == peak_freq)[0][0] for peak_freq in peak_freqs]
    peak_amplitude = amplitude[amplitude_index]

    peak_fig = {}

    if visualize:
        fig, ax = plot_peaks(lags, xcoef, freqs, power_norm, peak_freqs, peak_power_norm)
        peak_fig = {"fig": fig, "ax": ax}

    peaks = {"f": peak_freqs, "amplitude": peak_amplitude}

    return peaks, peak_fig


def _validate_data(y: np.ndarray):
    if not np.issubdtype(y.dtype, np.floating):
        raise TypeError("Data must be floating-point")

    if y.ndim != 1:
        raise TypeError(f"Signal data must have shape (samples,). Received shape={y.shape}")

    if np.isnan(y).any() or np.isinf(y).any():
        raise TypeError("Signal is not finite everywhere or contains one or more NaNs")

    if np.all(y[0] == y[1:]):
        raise ValueError("Ill-conditioned input array; contains only one unique value")


def lpc(y: np.ndarray, order: int) -> np.ndarray:
    """
    The code is adapted from https://github.com/librosa/librosa/blob/main/librosa/core/audio.py

    Librosa is licensed under the ISC License:

        Copyright (c) 2013--2017, librosa development team.

        Permission to use, copy, modify, and/or distribute this software for any
        purpose with or without fee is hereby granted, provided that the above
        copyright notice and this permission notice appear in all copies.

        THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
        WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
        MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
        ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
        WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
        ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
        OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

    Linear Prediction Coefficients via Burg's method

        This function applies Burg's method to estimate coefficients of a linear
        filter on ``y`` of order ``order``.  Burg's method is an extension to the
        Yule-Walker approach, which are both sometimes referred to as LPC parameter
        estimation by autocorrelation.

        It follows the description and implementation approach described in the
        introduction by Marple. [#]_  N.B. This paper describes a different method, which
        is not implemented here, but has been chosen for its clear explanation of
        Burg's technique in its introduction.

        .. [#] Larry Marple.
               A New Autoregressive Spectrum Analysis Algorithm.
               IEEE Transactions on Accoustics, Speech, and Signal Processing
               vol 28, no. 4, 1980.

    Args:
        y (np.ndarray): Uniformly sampled signal data.
        order (int): Order of the prediction polynomial. Hence number of LPC coefficients.
    Returns:
        np.ndarray: LPC coefficients.
    """

    if not isinstance(order, int) or order < 1:
        raise ValueError("order must be an integer > 0")

    _validate_data(y)

    return _lpc(y, order)


@njit
def _lpc(y: np.ndarray, order: int) -> np.ndarray:
    # This implementation follows the description of Burg's algorithm given in
    # section III of Marple's paper referenced in the docstring.
    #
    # We use the Levinson-Durbin recursion to compute AR coefficients for each
    # increasing model order by using those from the last. We maintain two
    # arrays and then flip them each time we increase the model order so that
    # we may use all the coefficients from the previous order while we compute
    # those for the new one. These two arrays hold ar_coeffs for order M and
    # order M-1.  (Corresponding to a_{M,k} and a_{M-1,k} in eqn 5)

    dtype = y.dtype.type
    ar_coeffs = np.zeros(order + 1, dtype=dtype)
    ar_coeffs[0] = dtype(1)
    ar_coeffs_prev = np.zeros(order + 1, dtype=dtype)
    ar_coeffs_prev[0] = dtype(1)

    # These two arrays hold the forward and backward prediction error. They
    # correspond to f_{M-1,k} and b_{M-1,k} in eqns 10, 11, 13 and 14 of
    # Marple. First they are used to compute the reflection coefficient at
    # order M from M-1 then are re-used as f_{M,k} and b_{M,k} for each
    # iteration of the below loop
    fwd_pred_error = y[1:]
    bwd_pred_error = y[:-1]

    # DEN_{M} from eqn 16 of Marple.
    den = np.dot(fwd_pred_error, fwd_pred_error) + np.dot(bwd_pred_error, bwd_pred_error)

    for i in range(order):
        if den <= 0:
            raise FloatingPointError("numerical error, input ill-conditioned?")

        # Eqn 15 of Marple, with fwd_pred_error and bwd_pred_error
        # corresponding to f_{M-1,k+1} and b{M-1,k} and the result as a_{M,M}
        # reflect_coeff = dtype(-2) * np.dot(bwd_pred_error, fwd_pred_error) / dtype(den)
        reflect_coeff = dtype(-2) * np.dot(bwd_pred_error, fwd_pred_error) / dtype(den)

        # Now we use the reflection coefficient and the AR coefficients from
        # the last model order to compute all of the AR coefficients for the
        # current one.  This is the Levinson-Durbin recursion described in
        # eqn 5.
        # Note 1: We don't have to care about complex conjugates as our signals
        # are all real-valued
        # Note 2: j counts 1..order+1, i-j+1 counts order..0
        # Note 3: The first element of ar_coeffs* is always 1, which copies in
        # the reflection coefficient at the end of the new AR coefficient array
        # after the preceding coefficients
        ar_coeffs_prev, ar_coeffs = ar_coeffs, ar_coeffs_prev
        for j in range(1, i + 2):
            ar_coeffs[j] = ar_coeffs_prev[j] + reflect_coeff * ar_coeffs_prev[i - j + 1]

        # Update the forward and backward prediction errors corresponding to
        # eqns 13 and 14.  We start with f_{M-1,k+1} and b_{M-1,k} and use them
        # to compute f_{M,k} and b_{M,k}
        fwd_pred_error_tmp = fwd_pred_error
        fwd_pred_error = fwd_pred_error + reflect_coeff * bwd_pred_error
        bwd_pred_error = bwd_pred_error + reflect_coeff * fwd_pred_error_tmp

        # SNIP - we are now done with order M and advance. M-1 <- M

        # Compute DEN_{M} using the recursion from eqn 17.
        #
        # reflect_coeff = a_{M-1,M-1}      (we have advanced M)
        # den =  DEN_{M-1}                 (rhs)
        # bwd_pred_error = b_{M-1,N-M+1}   (we have advanced M)
        # fwd_pred_error = f_{M-1,k}       (we have advanced M)
        # den <- DEN_{M}                   (lhs)
        #

        q = dtype(1) - reflect_coeff ** 2
        den = q * den - bwd_pred_error[-1] ** 2 - fwd_pred_error[0] ** 2

        # Shift up forward error.
        #
        # fwd_pred_error <- f_{M-1,k+1}
        # bwd_pred_error <- b_{M-1,k}
        #
        # N.B. We do this after computing the denominator using eqn 17 but
        # before using it in the numerator in eqn 15.
        fwd_pred_error = fwd_pred_error[1:]
        bwd_pred_error = bwd_pred_error[:-1]

    return ar_coeffs



from inso_toolbox.signal.butter import butterworth
from inso_toolbox.signal.chebyshev import chebyshev
from inso_toolbox.signal.interpolate import interpolate

from .sg import savitzky_golay

__all__ = ["interpolate", "chebyshev", "butterworth", "savitzky_golay"]



from .utils.converter_class import Converter
from .utils.unit_class import Unit
from .utils.units_class import Units

# Collect all unit groups inside this class instance
_UNITS = Units()

#################################################################################################
#   ADD YOUR OWN UNIT GROUPS BELOW                                                              #
#################################################################################################

# GUIDELINES:
# 1. GROUP NAMES MUST BE UNIQUE
# 2. ALL UNIT NAMES AND SYNONYMNS MUST BE UNIQUE
# 3. EACH UNIT GROUP MUST HAVE EXACTLY ONE DEFAULT (SI) UNIT
# 4. CONVERTER FUNCTIONS MUST HAVE ONE ARGUMENT AND RETURN A SCALAR
# (Automatic tests will be run to check that all unit groups and unit names/synonyms are unique.)


_UNITS.add_group(
    group_name="time",
    units=[
        Unit("sec", synonyms=["second", "seconds", "s"], is_si_unit=True),
        Unit("min", synonyms=["minute", "minutes"]),
        Unit("hr", synonyms=["hour", "hours"]),
    ],
    converters=[
        Converter("sec", "hr", lambda sec: sec / 60 / 60),
        Converter("hr", "sec", lambda hr: hr * 60 * 60),
        Converter("sec", "min", lambda sec: sec / 60),
        Converter("min", "sec", lambda min: min * 60),
        Converter("hr", "min", lambda sec: sec * 60),
        Converter("min", "hr", lambda min: min / 60),
    ],
)


_UNITS.add_group(
    group_name="length",
    units=[
        Unit("m", synonyms=["meter", "meters"], is_si_unit=True),
        Unit("mm", synonyms=["milli meter", "millimeter"]),
        Unit("cm", synonyms=["centi meter", "centimeter"]),
        Unit("ft", synonyms=["foot", "feet"]),
        Unit("in", synonyms=["inch", "inches"]),
    ],
    converters=[
        Converter("mm", "m", lambda mm: mm / 1000.0),
        Converter("cm", "m", lambda cm: cm / 100.0),
        Converter("ft", "m", lambda ft: ft * 0.3048),
        Converter("in", "m", lambda inch: inch * 0.0254),
        Converter("m", "mm", lambda m: m * 1000.0),
        Converter("m", "cm", lambda m: m * 100.0),
        Converter("m", "ft", lambda m: m / 0.3048),
        Converter("m", "in", lambda m: m / 0.0254),
    ],
)


_UNITS.add_group(
    group_name="mass",
    units=[Unit("kg", is_si_unit=True), Unit("lb", synonyms=["lbs"]), Unit("g")],
    converters=[
        Converter("kg", "lb", lambda kg: kg * 2.20462),
        Converter("lb", "kg", lambda lb: lb / 2.20462),
        Converter("g", "kg", lambda g: g / 1000),
        Converter("kg", "g", lambda kg: kg * 1000),
    ],
)


_UNITS.add_group(
    group_name="temperature",
    units=[
        Unit("k", synonyms=["kelvin"], is_si_unit=True),
        Unit("c", synonyms=["celsius", "°c"]),
        Unit("f", synonyms=["fahrenheit", "°f"]),
    ],
    converters=[
        Converter("k", "f", lambda k: (k - 273.15) * (9 / 5) + 32),
        Converter("k", "c", lambda k: k - 273.15),
        Converter("c", "k", lambda c: c + 273.15),
        Converter("c", "f", lambda c: c * (9 / 5) + 32),
        Converter("f", "k", lambda f: (f - 32) * (5 / 9) + 273.15),
        Converter("f", "c", lambda f: (f - 32) * (5 / 9)),
    ],
)


_UNITS.add_group(
    group_name="pressure",
    units=[Unit("pa", synonyms=["pascal"], is_si_unit=True), Unit("psi")],
    converters=[Converter("psi", "pa", lambda psi: psi * 6894.76), Converter("pa", "psi", lambda pa: pa / 6894.76)],
)


_UNITS.add_group(
    group_name="flow rate",
    units=[
        Unit("m3s", synonyms=["m3/s", "m³/s"], is_si_unit=True),
        Unit("m3min", synonyms=["m3/min"]),
        Unit("m3hrs", synonyms=["m3/hrs", "m3/hr", "m3hr"]),
        Unit("mmscfd", synonyms=["mmscf/d"]),
        Unit("usgpm"),
        Unit("usgpd"),
        Unit("mbpd", synonyms=["mbod"]),
        Unit("bpd", synonyms=["bbl/d", "bbld", "bopd", "bwpd"]),
    ],
    converters=[
        Converter("m3s", "m3min", lambda m3s: m3s * 60.0),
        Converter("m3s", "m3hrs", lambda m3s: m3s * 3600.0),
        Converter("m3s", "mmscfd", lambda m3s: m3s / 0.32774128),
        Converter("m3s", "mbpd", lambda m3s: m3s / 0.00184013),
        Converter("m3s", "usgpm", lambda m3s: m3s * 15850),
        Converter("m3s", "usgpd", lambda m3s: m3s * 15850 * 1440),
        Converter("m3s", "bpd", lambda m3s: m3s / 0.0000018401307),
        Converter("m3min", "m3s", lambda m3min: m3min / 60.0),
        Converter("m3hrs", "m3s", lambda m3hrs: m3hrs / 3600.0),
        Converter("mmscfd", "m3s", lambda mmscf: mmscf * 0.32774128),
        Converter("mbpd", "m3s", lambda mbpd: mbpd * 0.00184013),
        Converter("usgpm", "m3s", lambda usgpm: usgpm / 15850),
        Converter("usgpd", "m3s", lambda usgpd: usgpd / (15850 * 1440)),
        Converter("bpd", "m3s", lambda bpd: bpd * 0.0000018401307),
    ],
)


_UNITS.add_group(
    group_name="density",
    units=[
        Unit("kgm3", synonyms=["kg/m3", "kg/m³"], is_si_unit=True),
        Unit("lbft3", synonyms=["lb/ft3", "lb/ft³"]),
        Unit("sg", synonyms=["s.g.", "sp.gr."]),
    ],
    converters=[
        Converter("kgm3", "sg", lambda kgm3: kgm3 / 1000),
        Converter("sg", "kgm3", lambda sg: sg * 1000),
        Converter("kgm3", "lbft3", lambda kgm3: kgm3 / 16.01846),
        Converter("lbft3", "kgm3", lambda lbft3: lbft3 * 16.01846),
    ],
)


_UNITS.add_group(
    group_name="viscosity",
    units=[Unit("pas", synonyms=["pascalseconds"], is_si_unit=True), Unit("cp", synonyms=["centipoise"])],
    converters=[Converter("cp", "pas", lambda cp: cp * 0.001), Converter("pas", "cp", lambda pas: pas / 0.001)],
)


_UNITS.add_group(
    group_name="surface tension",
    units=[
        Unit("nm", synonyms=["n/m", "newton/m", "newton/meter"], is_si_unit=True),
        Unit("dyncm", synonyms=["dyn/cm", "dyne/cm", "dynes/cm"]),
    ],
    converters=[Converter("nm", "dyncm", lambda nm: nm * 1000), Converter("dyncm", "nm", lambda dyncm: dyncm / 1000)],
)


_UNITS.add_group(
    group_name="flow coefficient",
    units=[
        Unit("m3spa", synonyms=["(m3/s)/pa", "m3s/pa", "m3/spa"], is_si_unit=True),
        Unit("usgpmpsi", synonyms=["usgpm/psi"], is_si_unit=False),
        Unit("m3hbar", synonyms=["(m3/h)/bar", "m3h/bar", "m3/hbar"], is_si_unit=False),
    ],
    converters=[
        Converter("usgpmpsi", "m3hbar", lambda cv: cv * 0.865),
        Converter("usgpmpsi", "m3spa", lambda cv: cv * 3.2113 ** (-9)),
        Converter("m3hbar", "usgpmpsi", lambda cv: cv * 1.156),
        Converter("m3spa", "usgpmpsi", lambda cv: cv / 3.2113 ** (-9)),
    ],
)



import enum
import functools
from typing import Callable, Union

import numpy as np

from .darcy_friction import (
    _colebrook_dimless,
    _colebrook_fast_dimless,
    _colebrook_solve_dimless,
    _friction_factor_laminar_dimless,
    _haaland_dimless,
)

Numerical = Union[float, np.ndarray]


@enum.unique
class DarcyMethods(enum.Enum):
    """Collection of functions for calculating the darcy friction factor for turbulent flow, as a function of Reynolds
    number and effective roughness.  Members of this enumeration can be called and have the syntax described below. The
    four methods are different ways of solving the Colebrook-White equation, and are the following,

    See individual methods in ´inso_toolbox.physics.darcy_friction.py´ for more documentation.

    Members:
        * COLEBROOK: Uses the Lambert W function. Solves the equation exactly, but is relatively slow as the Lambert W
          function is solved iteratively.
        * COLEBROOK_SLOW: Solves the Colebrook-White equation using an iterative method.
        * COLEBROOK_FAST: An approximation to the Colebrook-White equation developed by D. Biberg[1].
        * HAALAND: The Haaland approximation to the Colebrook-White equation[2].

    References:
        * [1]: "Fast and Accurate Approximations for the Colebrook Equation", Dag Biberg,
          Journal of Fluids Engineering Copyright VC 2017 by ASME MARCH 2017, Vol. 139 / 031401-1,
          https://doi.org/10.1115/1.4034950
        * [2]: Haaland, SE (1983). "Simple and Explicit Formulas for the Friction Factor in Turbulent Flow". Journal of
          Fluids Engineering. 105 (1): 89–90. doi:10.1115/1.3240948.
    """

    COLEBROOK = functools.partial(_colebrook_dimless)
    COLEBROOK_SLOW = np.vectorize(_colebrook_solve_dimless)
    COLEBROOK_FAST = functools.partial(_colebrook_fast_dimless)
    HAALAND = functools.partial(_haaland_dimless)

    def __call__(self, rey_num, epsilon_r, **kwargs):
        """
        Calls the function represented by the DarcyMethods member value.

        Args:
            rey_num (float or np.ndarray): Reynolds number = rho * u * D / mu
            epsilon_r (float): Pipe's relative roughness, given as pipe effective
                roughness (epsilon) scaled with the pipe diameter (D) - epsilon_r = epsilon / D

        Returns:
            f_c (float or np.ndarray): darcy_friction_factor for turbulent flow.

        """
        return self.value(rey_num, epsilon_r, **kwargs)


def darcy_friction_factor_dimless(
    rey_num: Numerical,
    epsilon_r: Numerical,
    laminar_limit: float = 2300,
    turbulent_limit: float = 4000,
    method: Union[DarcyMethods, Callable] = None,
) -> Numerical:
    """
    Computes the Darcy friction factor for pipe flow, for both laminar and turbulent regimes. The transition regime
    between laminar and turbulent is given as   . The laminar-turbulent transition is given by the transition limits

    Args:
        - rey_num (float or np.ndarray): Reynolds number = rho*u*D/mu.
        - epsilon_r (float or np.ndarray): Pipe's relative roughness, given as pipe effective
          roughness (epsilon) scaled with the pipe diameter (D) - epsilon_r = epsilon / D
        - laminar_limit (float): Limit where lower Reynolds numbers give pure laminar flow.
        - turbulent_limit (float): Limit where higher Reynolds numbers give pure turbulent flow.
        - method (callable): The method to calculate the darcy friction factor in the turbulent regime.
          Possible arguments:

            * Any DarcyMethods member
            * other callables with the signature f(Union[float, np.ndarray], float) -> float.
    """
    if method is None:
        method = DarcyMethods.COLEBROOK_FAST
    if not callable(method):
        raise TypeError("Method must be a callable or part of the DarcyMethods Enum.")

    rey_num = np.asarray(rey_num)

    if laminar_limit > turbulent_limit:
        raise ValueError("Turbulent limit must be higher than laminar limit.")

    lam_mask = rey_num < laminar_limit
    turb_mask = rey_num > turbulent_limit
    both_mask = ~(lam_mask | turb_mask)

    res_lam = np.array([], dtype=np.float64)
    res_turb = res_lam.copy()
    res_both = res_lam.copy()

    if np.any(lam_mask):
        res_lam = _friction_factor_laminar_dimless(rey_num[lam_mask])

    if np.any(turb_mask):
        res_turb = method(rey_num[turb_mask], epsilon_r)

    if np.any(both_mask):
        res_both_lam = _friction_factor_laminar_dimless(rey_num[both_mask])
        res_both_turb = method(rey_num[both_mask], epsilon_r)
        laminar_fraction = (rey_num[both_mask] - laminar_limit) / (turbulent_limit - laminar_limit)
        res_both = res_both_turb * laminar_fraction + (1 - laminar_fraction) * res_both_lam

    return np.concatenate([res_lam, res_both, res_turb]).reshape(rey_num.shape)



from typing import List, Optional, Union

import pandas as pd

from inso_toolbox.utils import pipeline_function

from .units import _UNITS
from .utils.units_docstring import list_units_docstring


@pipeline_function
@list_units_docstring
def convert(
    data: Union[float, int, pd.DataFrame],
    from_unit: str,
    to_unit: Optional[str] = None,
    columns: Optional[List] = None,
    inplace: bool = False,
) -> Union[float, int, pd.DataFrame]:
    """
    Convert specified columns in data from given unit to another unit.
    
    Args:
        data (float, int, pd.DataFrame): Data to be converted.
        columns (List[str], optional): List of columns to convert.
        from_unit (str): Current unit of the specified columns.
        to_unit (str, optional): Specifies which unit the columns should be converted to.
            Default is to convert to SI units.
        inplace (bool, optional): Specifies whether to convert the data inplace or not.
    
    Returns:
        pd.DataFrame: Data with converted columns.
    """  # noqa
    f = _UNITS.get_converter(from_unit=from_unit, to_unit=to_unit)

    if isinstance(data, pd.DataFrame):
        if not inplace:
            data = data.copy(deep=True)

        data[columns] = f(data[columns])
    else:
        data = f(data)

    return data



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['inso_toolbox',
 'inso_toolbox.detectors',
 'inso_toolbox.detectors.utils',
 'inso_toolbox.geometry',
 'inso_toolbox.physics',
 'inso_toolbox.physics.dimensionless',
 'inso_toolbox.physics.dynamic_limit_sep',
 'inso_toolbox.physics.single_phase',
 'inso_toolbox.quality',
 'inso_toolbox.signal',
 'inso_toolbox.ts_utils',
 'inso_toolbox.units',
 'inso_toolbox.units.utils',
 'inso_toolbox.utils']

package_data = \
{'': ['*']}

install_requires = \
['cognite-sdk>=2.6.2,<3.0.0',
 'matplotlib>=3.3.0,<4.0.0',
 'numba>=0.52.0',
 'scipy>=1.4.1,<2.0.0']

setup_kwargs = {
    'name': 'cognite-inso-toolbox',
    'version': '3.0.2',
    'description': 'Inso Toolbox',
    'long_description': None,
    'author': 'cognite',
    'author_email': 'gustavo.zarruk@cognite.com, johannes.kolberg@cognite.com, nicolas.agnes@cognite.com, angela.atkinson@cognite.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



import warnings
from typing import Union

import pandas as pd

from inso_toolbox.utils.helpers import check_uniform, make_uniform


def drift_detector(
    data: Union[pd.DataFrame, pd.Series],
    long_term_interval: str,
    short_term_interval: str,
    std_threshold: float,
    resolution: str,
    detect: str = "both",
) -> pd.DataFrame:
    """
        Drift detection algorithm comparing the rolling average of the a given long and short term interval.
        The deviation between the short and long term average is considered significant if it is above the given
        threshold times the rolling standard deviation of the long term interval.


        Args:
            data (Union[pd.DataFrame, pd.Series]): Data to detect drift on. Index must be datetime and dataframe can
             only contain one column in addition to index.
            long_term_interval (str): Length of long term interval. Follows Pandas DateTime convention.
            short_term_interval (str): Length of short term interval. Follows Pandas DateTime convention.
            std_threshold (float): Parameter that determines the threshold for what is considered a significant change.
             The threshold is set to long term rolling standard deviation times the parameter std_threshold.
            resolution (str): Temporal resolution for resampling if the timeseries is not uniform. Follows Numpy
             DateTime convention.
            detect (str, optional): Parameter to determine if the model should detect significant decreases, increases
             or both (either "decrease", "increase" or "both").


        Returns:
            pd.DataFrame: Dataframe with a new column ``drift`` which indicates if the time series is drifting (1) or
            steady (0).


        Warning:
            The returned dataframe does not include the first datapoints which are within the length of the long term
            interval, as rolling average and standard deviation cannot be computed for these values.


        Example:
            Drift detection setting the long term to 8 days, short term to 12 hours, threshold to 3 times rolling long
            term std, granularity to 1 minute, and detecting direction to both.

            >>> data = drift_detector(data, long_term_interval='3d', short_term_interval='4h',
            >>>                                       std_threshold=3, resolution='1 min', detect='both')

        """

    # Only pd.Series and pd.DataFrame inputs are supported
    assert isinstance(data, pd.Series) or isinstance(
        data, pd.DataFrame
    ), "Only pd.Series and pd.DataFrame inputs are supported."

    # Only one column in DataFrame
    if isinstance(data, pd.DataFrame):
        assert len(data.columns) == 1, "Input data can only have one column."

    if isinstance(data, pd.Series):
        df = data.to_frame()
    else:
        df = data.copy()

    # Warn user if there are any missing values
    if df.isna().any(axis=None):
        warnings.warn("There are missing values present in the time series.", RuntimeWarning)

    column_name = df.columns[0]

    # Control the data is uniform, if not then resample to given resolution
    if not check_uniform(df):
        df = make_uniform(df, resolution=resolution, interpolation="linear")

    # Compute long term average and std
    df["long_term_average"] = df[column_name].rolling(long_term_interval).mean()
    df["long_term_stds"] = df[column_name].rolling(long_term_interval).std()

    # Compute short term average
    df["short_term_average"] = df[column_name].rolling(short_term_interval).mean()

    # Remove values within the first long term interval
    start_ts = df.index[0] + pd.Timedelta(long_term_interval)
    df = df.loc[start_ts:, :].copy()

    # initialize boolean drift column
    df["drift"] = [0] * df.count().max()

    if detect == "both" or detect == "increase":
        df.loc[df["short_term_average"] > df["long_term_average"] + std_threshold * df["long_term_stds"], "drift"] = 1

    if detect == "both" or detect == "decrease":
        df.loc[df["short_term_average"] < df["long_term_average"] - std_threshold * df["long_term_stds"], "drift"] = 1

    df = df.dropna()

    return df[[column_name, "drift"]]



def k_factor(gas_velocity: float, liquid_density: float, vapor_density: float):
    """
    Calculating the actual k-factor, aka system load factor or vapor load factor
    Ref: https://en.wikipedia.org/wiki/Souders%E2%80%93Brown_equation
    Ref: GPSA Engineering Data Book [Gas Processing] 12th ed p.7-12

    Args:
        gas_velocity (float): gas velocity, m/s
        liquid_density (float): liquid density, kg/m3
        vapor_density (float): vapor / gas density, kg/m3
    Returns:
        float: k-factor, m/s
    """
    k_factor = gas_velocity * (vapor_density / (liquid_density - vapor_density)) ** (0.5)
    return k_factor



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def steady_state(
    data: pd.DataFrame, ratio_lim: float = 2.5, avg_filter: float = 0.2, var_filter: float = 0.1, debug: bool = False
):
    """
    Steady state detector for a time series signal.

    Args:
        data (pd.DataFrame): Data to detect steady state on.
        ratio_lim (float, optional): Specifies the variance ratio threshold if it is in steady state or not.
            A variance greather than ``ratio_lim`` means it is in transient state.
        avg_filter (float, optional): Parameter that determines how many datapoints to use for calculating the
            average value of the time series.
        var_filter (float, optional): Parameter that determines how many datapoints to use for calculating the variance
            ratio.
        debug (bool, optional): Specifies whether to return a dataframe with columns used for intermediate calculations
            and plot analysis diagrams useful for determining the right parameter values.

    Returns:
        pd.DataFrame: Dataframe with a new column ``SS`` which indicates if the time series is steady (1) or transient
        (0).

    Warning:
        The returned dataframe has fewer rows because the initial datapoints are removed to conserve statistical
        indipendence.

    Hint:
        Larger ``avg_filter`` and ``var_filter`` values mean that fewer datapoints are involved in the analysis, which
        has a benefit of reducing the time for the identifier to catch up to a process change, reducing the average run
        length (ARL). Lower values undesirably increase the ARL to detection, but increase precision. With the default
        values only the previous 45 datapoints will be used.

    Example:

        To identify the optimal hyperparameter values for your data, make sure to run the function in a notebook with
        ``debug`` set to ``True`` as this will output a dataframe with intermediate calculations and plot various
        analysis diagrams.

            >>> # This will output a dataframe and plot analysyis diagrams
            >>> ss_data = steady_state(data, debug=True)

    Note:
        The implemented algorithm is taken from this paper: https://bit.ly/3fNfQZO
    """
    assert isinstance(data, pd.DataFrame), "Input data must be pd.DataFrame"
    assert len(data.columns) == 1, "Input data can only have one column."

    assert ratio_lim >= 0.0

    assert avg_filter >= 0.0 and avg_filter <= 1.0
    assert var_filter >= 0.0 and var_filter <= 1.0

    # Copy data
    df = data.copy()

    # df.columns = ["PV"]  # PV = Process Variable
    colname = data.columns[0]

    # Determine number of datapoints to be used to initializa values (set between 10 and 25)
    Ls = int(len(data) * 0.05)
    assert Ls >= 10, "There are too few datapoints to detect steady states."
    if Ls > 50:
        Ls = 50

    # Initialize filter and variance
    filt0 = df[colname].iloc[0:Ls].mean()
    var0 = df[colname].iloc[0:Ls].var()
    initial = df[colname].iloc[Ls - 1]

    # Remove first Ls df points to conserve statistical independence
    df.drop(df.index[0:Ls], inplace=True)

    # Create internal calculation columns
    df["_filt"] = 0
    df["_var1"] = 0
    df["_var2"] = 0

    # Set algorithm alpha values
    alpha1 = avg_filter
    alpha2, alpha3 = var_filter, var_filter

    # Initialize filter
    df.loc[df.index[0], "_filt"] = alpha1 * df[colname][0] + (1 - alpha1) * filt0

    # Initialize variance based on difference between df and filtered df
    df.loc[df.index[0], "_var1"] = alpha2 * (df[colname][0] - filt0) ** 2 + (1 - alpha2) * var0

    # Initialize diltered variance based sequential df differences
    df.loc[df.index[0], "_var2"] = alpha3 * (df[colname][0] - initial) ** 2 + (1 - alpha3) * var0

    # Calculate filter and variance values
    for i in range(1, len(df)):
        df.loc[df.index[i], "_filt"] = alpha1 * df[colname][i] + (1 - alpha1) * df._filt[i - 1]
        df.loc[df.index[i], "_var1"] = alpha2 * (df[colname][i] - df._filt[i - 1]) ** 2 + (1 - alpha2) * df._var1[i - 1]
        df.loc[df.index[i], "_var2"] = (
            alpha3 * (df[colname][i] - df[colname][i - 1]) ** 2 + (1 - alpha3) * df._var2[i - 1]
        )

    # Calcualte Variance ratio
    df["_ratio"] = (2 - alpha1) * df._var1 / df._var2

    # Filter df based on variance ration threshold (Rlim)
    st_colname = "SS"
    df[st_colname] = 1
    df["_steady"] = df[colname]
    df.loc[df._ratio > ratio_lim, st_colname] = 0
    df.loc[df._ratio > ratio_lim, "_steady"] = np.nan

    # Calculate rolling median from steady values
    df["_steady_ravg"] = df._steady.rolling("1h").median()
    df.loc[df._ratio > ratio_lim, "_steady_ravg"] = np.nan

    # Return df with data column and steady state column
    if not debug:
        return df[[colname, "SS"]]

    # If in debugging mode, return entire dataframe and plot output
    fig, [ax1, ax2, ax3] = plt.subplots(3, 1, figsize=(14, 12), tight_layout=True, sharex=True)

    # Plot time series and where steady state has been detected
    ax1.set_title("Steady State Detection Analysis", fontsize=14)
    ax1.plot(df[colname], "-", label=colname)
    ax1.plot(df._steady, "g-", label="steady state")
    ax1.plot(df._steady_ravg, "k-", linewidth=3, label="SS rolling avg")
    ax1.set_ylabel(colname)
    ax1.grid(which="both")
    ax1.legend()

    # Plot variance ratio
    ax2.plot(df._ratio, "k-")
    ax2.set_ylabel("Variance Ratio")
    ax2.grid(which="both")
    ax2.axhline(ratio_lim, color="r", linestyle="-", label=f"ratio_lim = {ratio_lim}")
    ax2.legend()

    # Plot steady state and transient state periods
    ax3.plot(df["SS"], "k-")
    ax3.set_xlabel("Date and Time")
    ax3.set_ylabel("Steady (1) - Transient (0)")
    ax3.grid(which="both")

    return df
